#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['subprocess', 'SubprocessMngr', 'run_command', 'timer',
           'signal_handler', 'run', 'PIPE', 'STDOUT']

from pyutilib.subprocess import GlobalData
import time
import signal
import os
import sys
import tempfile
import subprocess
from six import itervalues
from threading import Thread

_mswindows = sys.platform.startswith('win')

if _mswindows:
    import ctypes

# Note: on many python interpreters, WindowsError is only defined on
# Windows.  Since we want to trap it below, we will declare a local
# Exception type that is WindowsError if WindowsError is defined,
# otherwise it is a bogus Exception class.
try:
    WindowsError()
    _WindowsError = WindowsError
except NameError:

    class _WindowsError(Exception):
        pass


_peek_available = True
try:
    if _mswindows:
        from msvcrt import get_osfhandle
        from win32pipe import PeekNamedPipe
        from win32file import ReadFile
    else:
        from select import select
except:
    _peek_available = False

import pyutilib.services
from pyutilib.common import ApplicationError
from pyutilib.misc import quote_split

try:
    unicode
except:
    basestring = unicode = str

if sys.platform.startswith('java'):
    PIPE = None
    STDOUT = None
else:
    PIPE = subprocess.PIPE
    STDOUT = subprocess.STDOUT

if sys.version_info[0:2] >= (3, 0):

    def bytes_cast(x):
        if not x is None:
            if isinstance(x, basestring) is True:
                return x.encode()  # Encode string to bytes type
            else:
                # lets assume its of type bytes
                return x
        else:
            return x
else:
    bytes_cast = lambda x: x  # Do nothing

#
# Setup the timer
#
if sys.version_info >= (3,3):
    # perf_counter is guaranteed to be monotonic and the most accurate timer
    timer = time.perf_counter
else:
    # On old Pythons, clock() is more accurate than time() on Windows
    # (.35us vs 15ms), but time() is more accurate than clock() on Linux
    # (1ns vs 1us).  However, since we are only worring about process
    # timeouts here, we will stick with strictly monotonic timers
    timer = time.clock


def kill_process(process, sig=signal.SIGTERM, verbose=False):
    """
    Kill a process given a process ID
    """
    pid = process.pid
    if GlobalData.debug or verbose:
        print("Killing process %d with signal %d" % (pid, sig))
    if _mswindows:
        PROCESS_TERMINATE = 1
        handle = ctypes.windll.kernel32.OpenProcess(PROCESS_TERMINATE,
                                                    False, pid)
        ctypes.windll.kernel32.TerminateProcess(handle, -1)
        ctypes.windll.kernel32.CloseHandle(handle)
    else:
        #
        # Kill process and all its children
        #
        pgid = os.getpgid(pid)
        if pgid == -1:
            print("  ERROR: invalid pid %d" % pid)
            sys.exit(1)
        os.killpg(pgid, signal.SIGTERM)
        process.terminate()
        #
        # This is a hack.  The Popen.__del__ method references
        # the 'os' package, and when a process is interupted this
        # package is deleted before Popen.  I can't figure out why
        # _del_ is being called when Python closes down, though.  HOWEVER,
        # we can hard-ware Popen.__del__ to return immediately by telling it
        # that it did not create a child process!
        #
        if not GlobalData.current_process is None:
            GlobalData.current_process._child_created = False


GlobalData.current_process = None
GlobalData.pid = None
GlobalData.signal_handler_busy = False


#
# A signal handler that passes on the signal to the child process.
#
def verbose_signal_handler(signum, frame):
    c = frame.f_code
    print('  Signal handler called from ', c.co_filename, c.co_name,
          frame.f_lineno)
    print("  Waiting...",)
    signal_handler(signum, frame, True)


def signal_handler(signum, frame, verbose=False):
    if GlobalData.signal_handler_busy:
        print("")
        print("  Signal handler is busy.  Aborting.")
        sys.exit(-signum)
    if GlobalData.current_process is None:
        print("  Signal", signum, "recieved, but no process queued")
        print("  Exiting now")
        sys.exit(-signum)
    if GlobalData.current_process is not None and\
       GlobalData.current_process.pid is not None and\
       GlobalData.current_process.poll() is None:
        GlobalData.signal_handler_busy = True
        kill_process(GlobalData.current_process, signum)
        if verbose:
            print("  Signaled process", GlobalData.current_process.pid,
                  "with signal", signum)
        endtime = timer() + 1.0
        while timer() < endtime:
            status = GlobalData.current_process.poll()
            if status is None:
                break
            time.sleep(0.1)
        #GlobalData.current_process.wait()
        status = GlobalData.current_process.poll()
        if status is not None:
            GlobalData.signal_handler_busy = False
            if verbose:
                print("Done.")
            raise OSError("Interrupted by signal " + repr(signum))
        else:
            raise OSError("Problem terminating process" + repr(
                GlobalData.current_process.pid))
        GlobalData.current_process = None

    # Restore the original signal handlers (because the subprocess is
    # now defunct, and we shouldn't return here)
    for _sig in list(GlobalData.original_signal_handlers):
        signal.signal(_sig, GlobalData.original_signal_handlers.pop(_sig))
    # If there was originally a signal handler, then we should trigger
    # it, too
    orig_handler = signal.getsignal(signum)
    if hasattr(orig_handler, '__call__'):
        orig_handler(signum, frame)
    raise OSError("Interrupted by signal " + repr(signum))


#
# A function used to read in data from a shell command, and push it into a pipe.
#
def _stream_reader(args):
    unbuffer = args[0]
    stream = args[1]
    output = tuple(x for x in args[2:] if x is not None)

    def write(x):
        success = True
        for s in output:
            try:
                s.write(x)
            except ValueError:
                success = False
        return success

    def flush():
        for s in output:
            try:
                s.flush()
            except ValueError:
                pass

    raw_stderr = sys.__stderr__
    if raw_stderr is None:
        raw_stderr = sys.stderr
    try:
        encoding = stream.encoding
    except:
        encoding = None
    if encoding is None:
        try:
            encoding = raw_stderr.encoding
        except:
            pass
    if encoding is None:
        encoding = 'utf-8'

    buf = ""
    data = None

    while True:
        new_data = os.read(stream, 1)
        if not new_data:
            break
        if data:
            data += new_data
        else:
            data = new_data
        char = data.decode(encoding)
        if char.encode(encoding) != data:
            continue
        data = ""
        if unbuffer == 1:
            writeOK = write(char)
        buf += char
        if char[-1] != "\n":
            continue
        if unbuffer:
            if unbuffer != 1:
                writeOK = write(buf)
            flush()
        else:
            writeOK = write(buf)
        if writeOK:
            buf = ""
    writeOK = True
    if buf:
        writeOK &= write(buf)
    if data:
        writeOK &= write(data.decode(encoding))
    flush()
    if not writeOK and raw_stderr is not None:
        raw_stderr.write("""
ERROR: pyutilib.subprocess: output stream closed before all subprocess output
       was written to it.  The following was left in the subprocess buffer:
            '%s'
""" % (buf,))
        if data:
            raw_stderr.write(
                """The following undecoded unicode output was also present:
            '%s'
""" % (data,))


#
# A function used to read in data from two independent streams and push
# each to 1+ output pipes.  Managing this in a single thread allows our
# code output to be more predictible (and better formatted) when the
# output streams happen to point to the same place (e.g., file or
# terminal).  This function requires the ability to "peek" or "select"
# the next stream that is ready to be read from.
#
# For platforms that do not support select / peek, see the
# _pseudo_merged_reader.
#
def _merged_reader(*args):

    class StreamData(object):
        __slots__ = ('read', 'output', 'unbuffer', 'buf', 'data')

        def __init__(self, *args):
            if _mswindows:
                self.read = get_osfhandle(args[1])
            else:
                self.read = args[1]
            self.unbuffer = args[0]
            self.output = tuple(x for x in args[2:] if x is not None)
            self.buf = ""
            self.data = None

        def write(self, x):
            success = True
            for s in self.output:
                try:
                    s.write(x)
                except ValueError:
                    success = False
            return success

        def flush(self):
            for s in self.output:
                try:
                    s.flush()
                except ValueError:
                    pass

    raw_stderr = sys.__stderr__
    if raw_stderr is None:
        # There are cases, e.g., in Anaconda, where there is no stdout
        # for the original process because, for example, it was started
        # in a windowing environment.
        raw_stderr = sys.stderr
    try:
        encoding = stream.encoding
    except:
        encoding = None
    if encoding is None:
        try:
            encoding = raw_stderr.encoding
        except:
            pass
    if encoding is None:
        encoding = 'utf-8'

    streams = {}
    for s in args:
        tmp = StreamData(*s)
        streams[tmp.read] = tmp

    handles = sorted(streams.keys(), key=lambda x: -1 * streams[x].unbuffer)
    noop = []

    while handles:
        if _mswindows:
            new_data = None
            for h in handles:
                try:
                    numAvail = PeekNamedPipe(h, 0)[1]
                    if numAvail == 0:
                        continue
                    result, new_data = ReadFile(h, 1, None)
                except:
                    handles.remove(h)
                    new_data = None
            if new_data is None:
                continue
        else:
            h = select(handles, noop, noop)[0]
            if not h:
                break
            h = h[0]
            new_data = os.read(h, 1)
            if not new_data:
                handles.remove(h)
                continue
        s = streams[h]
        if s.data is None:
            s.data = new_data
        else:
            s.data += new_data
        char = s.data.decode(encoding)
        if char.encode(encoding) != s.data:
            continue
        s.data = None
        if s.unbuffer:
            writeOK = s.write(char)
        s.buf += char
        if char[-1] != "\n":
            continue
        if s.unbuffer:
            s.flush()
        else:
            writeOK = s.write(s.buf)
        if writeOK:
            s.buf = ""
    writeOK = True
    for s in itervalues(streams):
        if s.buf:
            writeOK &= s.write(s.buf)
        if s.data:
            writeOK &= s.write(s.data.decode(encoding))
        s.flush()
    if not writeOK and raw_stderr is not None:
        raw_stderr.write("""
ERROR: pyutilib.subprocess: output stream closed before all subprocess output
       was written to it.  The following was left in the subprocess buffer:
            '%s'
""" % (buf,))
        if data:
            raw_stderr.write(
                """The following undecoded unicode output was also present:
            '%s'
""" % (data,))


#
# A mock-up of the _merged_reader for platforms (or installations) that
# lack working select/peek implementations.  Note that this function
# does not guarantee determinism (and on many platforms is exceedingly
# nondeterministic).  However, it does change the flushing rules to
# better maintain output line integrity (at the cost of performance).
#
def _pseudo_merged_reader(*args):
    _threads = []
    for arg in args:
        _threads.append(Thread(target=_stream_reader, args=((2,) + arg[1:],)))
        _threads[-1].daemon = True
        _threads[-1].start()
    for th in _threads:
        th.join()


#
# Execute the command as a subprocess that we can send signals to.
# After this is finished, we can get the output from this command from
# the process.stdout file descriptor.
#
def run_command(cmd,
                outfile=None,
                cwd=None,
                ostream=None,
                stdin=None,
                stdout=None,
                stderr=None,
                valgrind=False,
                valgrind_log=None,
                valgrind_options=None,
                memmon=False,
                env=None,
                define_signal_handlers=None,
                debug=False,
                verbose=True,
                timelimit=None,
                tee=None,
                ignore_output=False,
                shell=False,
                thread_reader=None):
    #
    # Set the define_signal_handlers based on the global default flag.
    #
    if define_signal_handlers is None:
        define_signal_handlers = GlobalData.DEFINE_SIGNAL_HANDLERS_DEFAULT
    #
    # Move to the specified working directory
    #
    if cwd is not None:
        oldpwd = os.getcwd()
        os.chdir(cwd)

    cmd_type = type(cmd)
    if cmd_type is list:
        # make a private copy of the list
        _cmd = cmd[:]
    elif cmd_type is tuple:
        _cmd = list(cmd)
    else:
        _cmd = quote_split(cmd.strip())

    #
    # Setup memmoon
    #
    if memmon:
        memmon = pyutilib.services.registered_executable("memmon")
        if memmon is None:
            raise IOError("Unable to find the 'memmon' executable")
        _cmd.insert(0, memmon.get_path())
    #
    # Setup valgrind
    #
    if valgrind:
        #
        # The valgrind_log option specifies a logfile that is used to store
        # valgrind output.
        #
        valgrind_cmd = pyutilib.services.registered_executable("valgrind")
        if valgrind_cmd is None:
            raise IOError("Unable to find the 'valgrind' executable")
        valgrind_cmd = [valgrind_cmd.get_path()]
        if valgrind_options is None:
            valgrind_cmd.extend(
                ("-v", "--tool=memcheck", "--trace-children=yes"))
        elif type(valgrind_options) in (list, tuple):
            valgrind_cmd.extend(valgrind_options)
        else:
            valgrind_cmd.extend(quote_split(valgrind_options.strip()))
        if valgrind_log is not None:
            valgrind_cmd.append("--log-file-exactly=" + valgrind_log.strip())
        _cmd = valgrind_cmd + _cmd
    #
    # Redirect stdout and stderr
    #
    tmpfile = None
    if ostream is not None:
        stdout_arg = stderr_arg = ostream
        if outfile is not None or stdout is not None or stderr is not None:
            raise ValueError("subprocess.run_command(): ostream, outfile, and "
                             "{stdout, stderr} options are mutually exclusive")
        output = "Output printed to specified ostream"
    elif outfile is not None:
        stdout_arg = stderr_arg = open(outfile, "w")
        if stdout is not None or stderr is not None:
            raise ValueError("subprocess.run_command(): outfile and "
                             "{stdout, stderr} options are mutually exclusive")
        output = "Output printed to file '%s'" % outfile
    elif not (stdout is None and stderr is None):
        stdout_arg = stdout
        stderr_arg = stderr
        output = "Output printed to specified stdout and stderr streams"
    else:
        # Create a temporary file.  The mode is w+, which means that we
        #   can read and write.
        # NOTE: the default mode is w+b, but writing to the binary mode
        #   seems to cause problems in the _stream_reader function on Python
        #   3.x.
        stdout_arg = stderr_arg = tmpfile = tempfile.TemporaryFile(mode='w+')
        output = ""

    if stdout_arg is stderr_arg:
        try:
            if not tee or (not tee[0] and not tee[1]):
                stderr_arg = STDOUT
        except:
            pass

    #
    # Setup the default environment
    #
    if env is None:
        env = os.environ.copy()
    #
    # Setup signal handler
    #
    if define_signal_handlers:
        handler = verbose_signal_handler if verbose else signal_handler
        if sys.platform[0:3] != "win" and sys.platform[0:4] != 'java':
            GlobalData.original_signal_handlers[signal.SIGHUP] \
                = signal.signal(signal.SIGHUP, handler)
        GlobalData.original_signal_handlers[signal.SIGINT] \
            = signal.signal(signal.SIGINT, handler)
        GlobalData.original_signal_handlers[signal.SIGTERM] \
            = signal.signal(signal.SIGTERM, handler)
    rc = -1
    if debug:
        print("Executing command %s" % (_cmd,))
    try:
        try:
            simpleCase = not tee
            if stdout_arg is not None:
                stdout_arg.fileno()
            if stderr_arg is not None:
                stderr_arg.fileno()
        except:
            simpleCase = False

        out_th = []
        th = None
        GlobalData.signal_handler_busy = False
        if simpleCase:
            #
            # Redirect IO to the stdout_arg/stderr_arg files
            #
            process = SubprocessMngr(
                _cmd,
                stdin=stdin,
                stdout=stdout_arg,
                stderr=stderr_arg,
                env=env,
                shell=shell)
            GlobalData.current_process = process.process
            rc = process.wait(timelimit)
            GlobalData.current_process = None
        else:
            #
            # Aggressively wait for output from the process, and
            # send this to both the stdout/stdarg value, as well
            # as doing a normal 'print'
            #
            out_fd = []
            for fid in (0, 1):
                if fid == 0:
                    s, raw = stdout_arg, sys.stdout
                else:
                    s, raw = stderr_arg, sys.stderr
                try:
                    tee_fid = tee[fid]
                except:
                    tee_fid = tee
                if s is None or s is STDOUT:
                    out_fd.append(s)
                elif not tee_fid:
                    # This catches using StringIO as an output buffer:
                    # Python's subprocess requires the stream objects to
                    # have a "fileno()" attribute, which StringIO does
                    # not have.  We will mock things up by putting a
                    # pipe in between the subprocess and the StringIO
                    # buffer.  <sigh>
                    #
                    #if hasattr(s, 'fileno'):
                    #
                    # Update: in Python 3, StringIO declares a fileno()
                    # method, but that method throws an exception.  So,
                    # we can't just check for the attribute: we *must*
                    # call the method and see if we get an exception.
                    try:
                        s.fileno()
                        out_fd.append(s)
                    except:
                        r, w = os.pipe()
                        out_fd.append(w)
                        out_th.append(((fid, r, s), r, w))
                        #th = Thread(target=thread_reader, args=(r,None,s,fid))
                        #out_th.append((th, r, w))
                else:
                    r, w = os.pipe()
                    out_fd.append(w)
                    out_th.append(((fid, r, raw, s), r, w))
                    #th = Thread( target=thread_reader, args=(r,raw,s,fid) )
                    #out_th.append((th, r, w))
                #
            process = SubprocessMngr(
                _cmd,
                stdin=stdin,
                stdout=out_fd[0],
                stderr=out_fd[1],
                env=env,
                shell=shell)
            GlobalData.current_process = process.process
            GlobalData.signal_handler_busy = False
            #
            # Create a thread to read in stdout and stderr data
            #
            if out_th:
                if thread_reader is not None:
                    reader = thread_reader
                elif len(out_th) == 1:
                    reader = _stream_reader
                elif _peek_available:
                    reader = _merged_reader
                else:
                    reader = _pseudo_merged_reader
                th = Thread(target=reader, args=[x[0] for x in out_th])
                th.daemon = True
                th.start()
            #
            # Wait for process to finish
            #
            rc = process.wait(timelimit)
            GlobalData.current_process = None
            out_fd = None

    except _WindowsError:
        err = sys.exc_info()[1]
        raise ApplicationError(
            "Could not execute the command: '%s'\n\tError message: %s" %
            (' '.join(_cmd), err))
    except OSError:
        #
        # Ignore IOErrors, which are caused by interupts
        #
        pass
    finally:
        # restore the previous signal handlers, if necessary
        for _sig in list(GlobalData.original_signal_handlers):
            signal.signal(_sig, GlobalData.original_signal_handlers.pop(_sig))

    #
    # Flush stdout/stderr. Some platforms (notably Matlab, which
    # replaces stdout with a MexPrinter) have stdout/stderr that do not
    # implement flush()  See https://github.com/Pyomo/pyomo/issues/156
    #
    try:
        sys.stdout.flush()
    except AttributeError:
        pass
    try:
        sys.stderr.flush()
    except AttributeError:
        pass

    if out_th:
        #
        # 'Closing' the PIPE to send EOF to the reader.
        #
        for p in out_th:
            os.close(p[2])
        if th is not None:
            # Note, there is a condition where the subprocess can die
            # very quickly (raising an OSError) before the reader
            # threads have a chance to be set up.  Testing for None
            # avoids joining a thread that doesn't exist.
            th.join()
        for p in out_th:
            os.close(p[1])
        if th is not None:
            del th
    if outfile is not None:
        stdout_arg.close()
    elif tmpfile is not None and not ignore_output:
        tmpfile.seek(0)
        output = "".join(tmpfile.readlines())
        tmpfile.close()
    #
    # Move back from the specified working directory
    #
    if cwd is not None:
        os.chdir(oldpwd)
    #
    # Return the output
    #
    return [rc, output]

# Create an alias for run_command
run = run_command


class SubprocessMngr(object):

    def __init__(self,
                 cmd,
                 stdin=None,
                 stdout=None,
                 stderr=None,
                 env=None,
                 bufsize=0,
                 shell=False):
        """
        Setup and launch a subprocess
        """
        self.process = None
        #
        # By default, stderr is mapped to stdout
        #
        #if stderr is None:
        #    stderr=subprocess.STDOUT

        self.stdin = stdin
        if stdin is None:
            stdin_arg = None
        else:
            stdin_arg = subprocess.PIPE
        #
        # We would *really* like to deal with commands in execve form
        #
        if type(cmd) not in (list, tuple):
            cmd = quote_split(cmd.strip())
        #
        # Launch subprocess using a subprocess.Popen object
        #
        if _mswindows:
            #
            # Launch without console on MSWindows
            #
            startupinfo = subprocess.STARTUPINFO()
            #startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            self.process = subprocess.Popen(
                cmd,
                stdin=stdin_arg,
                stdout=stdout,
                stderr=stderr,
                startupinfo=startupinfo,
                env=env,
                bufsize=bufsize,
                shell=shell)
        elif getattr(subprocess, 'jython', False):
            #
            # Launch from Jython
            #
            self.process = subprocess.Popen(
                cmd,
                stdin=stdin_arg,
                stdout=stdout,
                stderr=stderr,
                env=env,
                bufsize=bufsize,
                shell=shell)
        else:
            #
            # Launch on *nix
            #
            self.process = subprocess.Popen(
                cmd,
                stdin=stdin_arg,
                stdout=stdout,
                stderr=stderr,
                preexec_fn=os.setsid,
                env=env,
                bufsize=bufsize,
                shell=shell)

    def X__del__(self):
        """
        Cleanup temporary file descriptor and delete that file.
        """

        if False and self.process is not None:
            try:
                if self.process.poll() is None:
                    self.kill()
            except OSError:
                #
                # It should be OK to ignore this exception.  Although poll() returns
                # None when the process is still active, there is a race condition
                # here.  Between running poll() and running kill() the process
                # may have terminated.
                #
                pass
        if self.process is not None:
            try:
                del self.process
            except:
                pass
        self.process = None

    def wait(self, timelimit=None):
        """
        Wait for the subprocess to terminate.  Terminate if a specified
        timelimit has passed.
        """
        if timelimit is None:
            # *Py3k: bytes_cast does no conversion for python 2.*, casts to bytes for 3.*
            self.process.communicate(input=bytes_cast(self.stdin))
            return self.process.returncode
        else:
            #
            # Wait timelimit seconds and then force a termination
            #
            # Sleep every 1/10th of a second to avoid wasting CPU time
            #
            if timelimit <= 0:
                raise ValueError("'timeout' must be a positive number")
            endtime = timer() + timelimit

            # This might be dangerous: we *could* deadlock if the input
            # is large...
            if self.stdin is not None:
                # *Py3k: bytes_cast does no conversion for python 2.*, casts to bytes for 3.*
                self.process.stdin.write(bytes_cast(self.stdin))

            while timer() < endtime:
                status = self.process.poll()
                if status is not None:
                    return status
                time.sleep(0.1)
            #
            # Check one last time before killing the process
            #
            status = self.process.poll()
            if status is not None:
                return status
            #
            # If we're here, then kill the process and return an error
            # returncode.
            #
            try:
                self.kill()
                return -1
            except OSError:
                #
                # The process may have stopped before we called 'kill()'
                # so check the status one last time.
                #
                status = self.process.poll()
                if status is not None:
                    return status
                else:
                    raise OSError("Could not kill process " + repr(
                        self.process.pid))

    def stdout(self):
        return self.process.stdout

    def send_signal(self, sig):
        """
        Send a signal to a subprocess
        """
        os.signal(self.process, sig)

    def kill(self, sig=signal.SIGTERM):
        """
        Kill the subprocess and its children
        """
        kill_process(self.process, sig)
        self.process.terminate()
        self.process.wait()
        del self.process
        self.process = None


if __name__ == "__main__":  #pragma:nocover
    #GlobalData.debug=True
    print("Z")
    stime = timer()
    foo = run_command("./dummy", tee=True, timelimit=10)
    print("A")
    print("Ran for " + repr(timer() - stime) + " seconds")
    print(foo)
    sys.exit(0)

    if not _mswindows:
        foo = SubprocessMngr("ls *py")
        foo.wait()
        print("")

        foo = SubprocessMngr("ls *py", stdout=subprocess.PIPE)
        foo.wait()
        for line in foo.process.stdout:
            print(line,)
        print("")
        foo = None

        [rc, output] = run_command("ls *py")
        print(output)
        print("")

        [rc, output] = run_command("ls *py", outfile="tmp")
        INPUT = open("tmp", 'r')
        for line in INPUT:
            print(line,)
        INPUT.close()
        print("")

        print("X")
        [rc, output] = run_command(
            "python -c \"while True: print '.'\"", timelimit=2)
        print("Y")
        #[rc,output] = run_command("python -c \"while True: print '.'\"")
        [rc, output] = run_command(
            "python -c \"while True: print '.'\"", verbose=False)
        print("Y-end")
    else:
        foo = SubprocessMngr("cmd /C \"dir\"")
        foo.wait()
        print("")

    print("Z")
    stime = timer()
    foo = run_command("python -c \"while True: pass\"", timelimit=10)
    print("A")
    print("Ran for " + repr(timer() - stime) + " seconds")

pyutilib.services.register_executable("valgrind")
pyutilib.services.register_executable("memmon")



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['infinity', 'nan', 'is_nan', 'is_finite']
"""
Definitions of mathematical constants
"""

import math


""" Definition of infinity """
infinity = float('inf')
""" Definition of NaN """
nan = infinity / infinity

def is_nan(x):
    """
    Returns true if the argument is a float and it does not equal itself
    """
    try:
        return math.isnan(x)
    except TypeError:
        return False

def is_finite(x):
    """
    Returns true if the argument is a float or int and it is not infinite or NaN
    """
    try:
        return not math.isinf(x)
    except TypeError:
        return False



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#
from . import core

#
# Import the 'pyutilib.component' plugins
#
core.PluginGlobals.add_env("pca")

from . import config
from . import executables
from . import loader
from . import app

#
# Remove the "pca" environment as the default
#
core.PluginGlobals.pop_env()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""Definitions for workflow task objects."""

__all__ = ['Task', 'EmptyTask', 'Component', 'Port', 'Ports', 'InputPorts',
           'OutputPorts', 'Connector', 'DirectConnector']

import argparse
import pprint
import weakref
from pyutilib.misc import Options
from pyutilib.workflow import globals


class Task(object):
    """
    A Task object represents a single action in a workflow.
    """

    def __init__(self, id=None, name=None, parser=None):
        """Constructor."""
        if not id is None:
            self.id = id
        else:
            self.id = globals.unique_id()
        if name is None:
            self.name = "Task" + str(self.id)
        else:
            self.name = name
        self.inputs = InputPorts(self)
        self.inputs.set_name(self.name + "-inputs")
        self.outputs = OutputPorts(self)
        self.outputs.set_name(self.name + "-outputs")
        self._resources = {}
        self._predecessors = []
        self._create_parser(parser)
        self.input_controls = InputPorts(self)
        self.input_controls.set_name(self.name + '-input-controls')
        self.output_controls = OutputPorts(self)
        self.output_controls.set_name(self.name + '-output-controls')
        self.debug = False

    def add_resource(self, resource):
        """Add a resource that is required for this task to execute."""
        self._resources[resource.name] = resource

    def resource(self, name):
        """Return the specified resource object."""
        return self._resources[name]

    def next_tasks(self):
        """Return the set of tasks that succeed this task in the workflow."""
        return set(t.to_port.task()
                   for name in self.outputs
                   for t in self.outputs[name].output_connections) | set(
                       t.to_port.task()
                       for name in self.output_controls
                       for t in self.output_controls[name].output_connections)

    def prev_tasks(self):
        """Return the set of tasks that precede this task in the workflow."""
        return set([task
                    for name in self.inputs
                    for task in self.inputs[name].from_tasks()
                    if task.id != NoTask.id]) | set(
                        task for task in self._predecessors) | set(
                            [task
                             for name in self.input_controls
                             for task in self.input_controls[name].from_tasks()
                             if task.id != NoTask.id])

    def next_task_ids(self):
        """Return the set of ids for tasks that succeed this task in the workflow."""
        return set(task.id for task in self.next_tasks())

    def prev_task_ids(self):
        """Return the set of ids for tasks that precede this task in the workflow."""
        return set(task.id for task in self.prev_tasks())

    def execute(self, debug=False):
        """Execute this task."""
        raise ValueError(
            "There is no default execution for an abstract Task object! Task=%s"
            % self._name())  #pragma:nocover

    def busy_resources(self):
        """Return the list of resources that this task is waiting for."""
        return [name for name in self._resources
                if not self._resources[name].available()]

    def ready(self):
        if self.busy():
            return False
        for name in self.inputs:
            #print "XYZ",self.name, name, self.inputs[name].ready(),self.inputs[name]._ready
            #for connection in self.inputs[name].input_connections:
            #print "XYZ",self.name, name,connection.from_port._ready, connection.ready(), len(connection.from_port.input_connections), connection.from_port.task.name
            if not self.inputs[name].ready():
                #print "FALSE - input", name
                #print self.inputs[name]
                return False
        for name in self.input_controls:
            if not self.input_controls[name].ready():
                #print "FALSE - control", name
                return False
        return True

    def busy(self):
        """Return the list of resources that this task is waiting for."""
        return len(self.busy_resources())

    def __call__(self, *options, **kwds):
        """Setup inputs and output parameters and execute this task.

        Copy the inputs into this Task's dictionary, then execute the task, then copy
        the outputs out of the dictionary.
        """
        self._call_init(*options, **kwds)
        self.execute()
        return self._call_fini(*options, **kwds)

    def _call_init(self, *options, **kwds):
        self._call_start()
        busy = self.busy_resources()
        if len(busy) > 0:
            raise IOError("Cannot execute task %s.  Busy resources: %s" %
                          (self.name, str(busy)))
        # Set inputs
        for opt in options:
            self._set_inputs(opt)
        self._set_inputs(kwds)
        #
        for name, res in self._resources.items():
            res.lock()
        for i in self.outputs:
            #print "z",i,getattr(self.outputs,i).get_value()
            setattr(self, i, None)
        for i in self.inputs:
            #print "OIUOX",i,self.inputs[i].get_value(),str(self.inputs[i])
            # TODO: validate that non-optional inputs have a value other than None
            self.inputs[i].compute_value()
            setattr(self, i, self.inputs[i].get_value())

    def _call_fini(self, *options, **kwds):
        for i in self.outputs:
            #print "Z",i,getattr(self.outputs,i).get_value()
            # TODO: validate that non-optional outputs have a value other than None
            self.outputs[i].set_value(getattr(self, i))

        for name, res in self._resources.items():
            res.unlock()
        self._call_finish()
        self.set_ready()
        #
        opt = Options()
        for i in self.outputs:
            setattr(opt, i, getattr(self.outputs, i).get_value())
        return opt

    def set_options(self, args):
        """Use a list of command-line options to initialize this task."""
        [self.options, args] = self._parser.parse_known_args(args)
        tmp = {}
        for action in self._parser._actions:
            try:
                val = getattr(self.options, action.dest)
                tmp[action.dest] = val
            except:
                pass
        self._set_inputs(tmp)

    def _call_start(self):
        """This method is executed when the task is started."""
        pass

    def _call_finish(self):
        """This method is executed when the task is finished."""
        pass

    def _set_inputs(self, options):
        """Set the inputs from a dictionary."""
        for key in options:
            self.inputs[key].set_value(options[key])

    def set_arguments(self, parser=None):
        if parser is None:
            return
        for arg in self._parser_arg:
            args = arg[0]
            kwargs = arg[1]
            self._parser.add_argument(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        self._parser_arg.append([args, kwargs])
        self._parser.add_argument(*args, **kwargs)

    def _create_parser(self, parser=None):
        """Create the OptionParser object and populate it with option groups."""
        if parser is None:
            self._parser = argparse.ArgumentParser()
        else:
            self._parser = parser
        self._parser_arg = []
        self._parser_group = {}
        self._create_parser_groups()
        for key in self._parser_group:
            self._parser.add_argument_group(self._parser_group[key])

    def _create_parser_groups(self):
        """This method is called by the _create_parser method to setup the
        parser groups that are registered for this task."""

    def _repn_(self):
        tmp = {}
        tmp['A_TYPE'] = 'Task'
        tmp['Name'] = self.name
        tmp['Id'] = self.id
        tmp['Inputs'] = self.inputs._repn_()
        tmp['Outputs'] = self.outputs._repn_()
        tmp['InputControls'] = self.input_controls._repn_()
        tmp['OutputControls'] = self.output_controls._repn_()
        return tmp

    #def __repr__(self):
    #"""Return a string representation for this task."""
    #return pprint.pformat(self._repn_(), 2)

    def __str__(self):
        """Return a string representation for this task."""
        return pprint.pformat(self._repn_(), 2)

    def _name(self):
        return "%s prev: %s next: %s resources: %s" % (
            str(self.name), str(sorted(list(self.prev_task_ids()))),
            str(sorted(list(self.next_task_ids()))),
            str(sorted(self._resources.keys())))

    def reset(self):
        #print "RESETING "+self.name
        for i in self.outputs:
            self.outputs[i].reset()
        for i in self.output_controls:
            self.output_controls[i].reset()

    def set_ready(self):
        for i in self.outputs:
            self.outputs[i].set_ready()


class Component(Task):
    """
    Alias for the Task class.
    """

    def __init__(self, *args, **kwds):
        """Constructor."""
        Task.__init__(self, *args, **kwds)  #pragma:nocover


class EmptyTask(Task):

    def __call__(self, *args, **kwds):
        """Empty task execution."""

    #
    # An empty task maintains no option parser
    #

    def set_options(self, args):
        """Empty task initialization."""

    def set_arguments(self, *args, **kwds):  #pragma:nocover
        raise NotImplementedError  #pragma:nocover

    def add_argument(self, *args, **kwds):  #pragma:nocover
        raise NotImplementedError  #pragma:nocover

    def _create_parser(self, *args, **kwds):
        self._parser = None
        self._parser_arg = []
        self._parser_group = {}

    def _create_parser_groups(self):  #pragma:nocover
        raise NotImplementedError  #pragma:nocover


def define_connection(cls, from_port, to_port):
    """Define a connection by constructing the specified class."""
    # TODO: Generate a warning if required input is connect to an optional output

    #
    # Raise an exception if the port action is store and there already exists a connection.
    #
    if to_port.action == 'store' and len(to_port.input_connections) == 1:
        raise ValueError(
            "Cannot connect to task %s port %s from task %s port %s. This port is already connected from task %s port %s"
            %
            (to_port.task().name, to_port.name, from_port.task().name,
             from_port.name, to_port.input_connections[0].from_port.task().name,
             to_port.input_connections[0].from_port.name))
    #
    #print 'connecting',from_port.task.id, to_port.task.id
    connector = cls(from_port=from_port, to_port=to_port)
    to_port.input_connections.append(connector)
    from_port.output_connections.append(connector)


class Port(object):
    """A class that represents an input or output port on a task."""

    def __init__(self,
                 name,
                 task,
                 optional=False,
                 value=None,
                 action=None,
                 constant=False,
                 default=None,
                 doc=None):
        """Constructor."""
        self.name = name
        # tasks are stored as weak refs, to prevent issues with cyclic dependencies and the garbage collector.
        self.task = weakref.ref(task)
        if action is None:
            self.action = 'store'
        else:
            self.action = action
        self.optional = optional
        self.constant = constant
        self.input_connections = []
        self.output_connections = []
        self.set_value(value)
        self._ready = False
        self.default = default
        self.doc = doc

    def reset(self):
        self._ready = False
        self.value = None

    def connect(self, port):
        """Define a connection with the specified port."""
        define_connection(DirectConnector, from_port=port, to_port=self)

    def from_tasks(self):
        """Return the id of the preceding task."""
        return [c.from_port.task() for c in self.input_connections
                if c.from_port.task() != None]

    def get_value(self):
        """Get the value of this port."""
        if self.value is None:
            return self.default
        return self.value

    def set_value(self, value):
        """Set the value of this port."""
        self.value = value

    def compute_value(self):
        """Compute the value from the input connections."""
        #print 'X',self.name, self.action, len(self.input_connections)
        if self.action == 'store':
            if len(self.input_connections) == 1:
                val = self.input_connections[0].get_value()
                if not val is None:
                    self.value = val

        elif self.action == 'store_any':
            for connection in self.input_connections:
                if not connection.ready():
                    continue
                val = connection.get_value()
                if not val is None:
                    self.value = val
                    break

        elif self.action in ['append', 'append_any']:
            tmp = []
            for connection in self.input_connections:
                val = connection.get_value()
                if not val is None:
                    tmp.append(val)
            if len(tmp) > 0:
                self.value = tmp

        elif self.action in ['map', 'map_any']:
            tmp = {}
            for connection in self.input_connections:
                val = connection.get_value()
                if not val is None:
                    tmp[connection.from_port.task().id] = val
            if len(tmp) > 0:
                self.value = tmp

        self.validate()

    def validate(self):
        if self.action in ['store', 'store_any']:
            if not self.optional and self.get_value() is None:
                raise ValueError(
                    "Task %s Port %s requires a nontrivial value.  Value specified is None."
                    % (str(self.task().id), self.name))
        #
        elif self.action in ['append', 'append_any']:
            if not self.optional and self.get_value() is None:
                raise ValueError(
                    "Task %s Port %s requires a nontrivial value.  All input connections have value None."
                    % (str(self.task().id), self.name))
        #
        elif self.action in ['map', 'map_any']:
            if not self.optional and self.get_value() is None:
                raise ValueError(
                    "Task %s Port %s requires a nontrivial value.  All input connections have value None."
                    % (str(self.task().id), self.name))

    def _repn_(self):
        tmp = {}
        tmp['A_TYPE'] = 'Port'
        tmp['Name'] = self.name
        tmp['Task'] = str(self.task().id)
        tmp['Optional'] = str(self.optional)
        tmp['Constant'] = str(self.constant)
        tmp['Value'] = str(self.value)
        tmp['Ready'] = str(self.ready())
        tmp['Action'] = self.action
        tmp['Connections'] = {}
        tmp['Connections']['Inputs'] = []
        for c in self.input_connections:
            tmp['Connections']['Inputs'].append(repr(c))
        tmp['Connections']['Outputs'] = []
        for c in self.output_connections:
            tmp['Connections']['Outputs'].append(repr(c))
        return tmp

    def __repr__(self):
        return str(self)  #pragma:nocover

    def __str__(self, indentation=""):
        return pprint.pformat(self._repn_(), 2)  #pragma:nocover

    def ready(self):
        if not self.get_value() is None:
            return True
        if len(self.input_connections) == 0:
            return self._ready
        if self.action == 'store':
            return self.input_connections[0].ready()
        elif self.action in ['append_any', 'map_any', 'store_any']:
            for connection in self.input_connections:
                if connection.ready():
                    return True
            return False
        elif self.action in ['append', 'map']:
            for connection in self.input_connections:
                if not connection.ready():
                    return False
            return True
        #
        # We should never get here.
        #
        raise IOError(
            "WARNING: unknown action: " + self.action)  #pragma:nocover

    def set_ready(self):
        self._ready = True


class Ports(dict):
    """A class that specifies a set of ports."""

    def __init__(self, task):
        """Constructor."""
        self._name_ = 'Ports'
        # tasks are stored as weak refs, to prevent issues with cyclic dependencies and the garbage collector.
        self._task = weakref.ref(task)
        self._inputs = False
        self._outputs = False

    def set_name(self, name):
        """Set the name of this class instance."""
        self._name_ = name

    def declare(self,
                name,
                optional=False,
                action=None,
                constant=False,
                default=None,
                doc=None):
        """Declare a port."""
        port = Port(
            name,
            self._task(),
            optional=optional,
            action=action,
            constant=constant,
            default=default,
            doc=doc)
        setattr(self, name, port)
        return port

    def __setitem__(self, name, val):
        """Overload this operator to set an attribute with the specified name."""
        self.__setattr__(name, val)

    def __getitem__(self, name):
        """Overload this operator to get the attribute with the specified name."""
        return self.__getattr__(name)

    def __setattr__(self, name, val):
        """Overload this operator to setup a connection."""
        #
        # Directly set attributes whose names begin with '_'
        #
        if name[0] == '_':
            self.__dict__[name] = val
            return
        #
        # Declare a port
        #
        if not name in self.__dict__:
            if not isinstance(val, Port):
                raise TypeError(
                    "Error declaring port '%s' without a Port object" % name)
            dict.__setitem__(self, name, val)
            self.__dict__[name] = val
        #
        # Create a connection to another port
        #
        else:
            if not isinstance(val, Port):
                self.__dict__[name].set_value(val)
            else:
                self.__dict__[name].connect(val)

    def __getattr__(self, name):
        """Overload this operator to setup a connection."""
        try:
            return self.__dict__.__getitem__(name)
        except:
            raise AttributeError("Unknown attribute '%s'" % name)

    def _repn_(self):
        tmp = {}
        tmp['A_TYPE'] = 'Port'
        for k, v in self.__dict__.items():
            if not k.startswith("_"):
                tmp[k] = v._repn_()
        tmp['Name'] = self._name_
        if self._inputs:
            tmp['Mode'] = 'inputs'
        if self._outputs:
            tmp['Mode'] = 'outputs'
        # JPW: a port can't really exist independently of a task - we should check 
        #      this in the constructor.
        if self._task is not None:
            tmp['Owner'] = self._task()._name()
        return tmp

    def __repr__(self):  #pragma:nocover
        """Return a string representation of these connections."""
        attrs = sorted("%s = %r" % (k, v) for k, v in self.__dict__.items()
                       if not k.startswith("_"))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(attrs))

    def __str__(self, nesting=1, indent='', print_name=True):
        return pprint.pformat(self._repn_(), 2, width=150)


class InputPorts(Ports):
    """A class that is used to manage set of input ports."""

    def __init__(self, task):
        """Constructor."""
        Ports.__init__(self, task)
        self._inputs = True


class OutputPorts(Ports):
    """A class that is used to manage set of output ports."""

    def __init__(self, task):
        """Constructor."""
        Ports.__init__(self, task)
        self._outputs = True


class Connector(object):

    def __init__(self, from_port=None, to_port=None):
        """Constructor."""
        if from_port is None:
            self.from_port = NoTask  #pragma:nocover
        else:
            self.from_port = from_port
        if to_port is None:
            self.to_port = NoTask  #pragma:nocover
        else:
            self.to_port = to_port

    def get_value(self):
        raise ValueError(
            "There is no value to get in an abstract Connector object!")  #pragma:nocover

    def ready(self):
        return self.from_port.ready()

    def __repr__(self):
        """Return a string representation for this connection."""
        return str(self)

    def __str__(self):
        """Return a string representation for this connection."""
        return "%s: from=(%s) to=(%s) %s" % (
            str(self.__class__.__name__), str(self.from_port.task().id),
            str(self.to_port.task().id), self.ready())


class DirectConnector(Connector):

    def __init__(self, from_port=None, to_port=None):
        Connector.__init__(self, from_port=from_port, to_port=to_port)

    def get_value(self):
        if self.ready():
            return self.from_port.get_value()
        return None

# A task instance that represents no task.
NoTask = EmptyTask(id=0)



#
# Plugin load tests, with the sys.path loader disabled.
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest
import pyutilib.subprocess

try:
    import yaml
    yaml_available = True
except ImportError:
    yaml_available = False


class Test(unittest.TestCase):

    def test_load1(self):
        pyutilib.subprocess.run(
            [sys.executable, currdir + os.sep + "load1.py", currdir, "json"])
        self.assertMatchesJsonBaseline(currdir + "load1.out",
                                       currdir + "load1.jsn")
        if yaml_available:
            pyutilib.subprocess.run([sys.executable, currdir + os.sep +
                                     "load1.py", currdir, "yaml"])
            self.assertMatchesYamlBaseline(currdir + "load1.out",
                                           currdir + "load1.yml")

    def test_load1a(self):
        pyutilib.subprocess.run(
            [sys.executable, currdir + os.sep + "load1a.py", currdir, "json"])
        self.assertMatchesJsonBaseline(currdir + "load1a.out",
                                       currdir + "load1a.jsn")
        if yaml_available:
            pyutilib.subprocess.run([sys.executable, currdir + os.sep +
                                     "load1a.py", currdir, "yaml"])
            self.assertMatchesYamlBaseline(currdir + "load1a.out",
                                           currdir + "load1a.yml")

    def test_load2(self):
        pyutilib.subprocess.run(
            [sys.executable, currdir + os.sep + "load2.py", currdir, "json"])
        self.assertMatchesJsonBaseline(currdir + "load2.out",
                                       currdir + "load2.jsn")
        if yaml_available:
            pyutilib.subprocess.run([sys.executable, currdir + os.sep +
                                     "load2.py", currdir, "yaml"])
            self.assertMatchesYamlBaseline(currdir + "load2.out",
                                           currdir + "load2.yml")

    def test_load2a(self):
        pyutilib.subprocess.run(
            [sys.executable, currdir + os.sep + "load2a.py", currdir, "json"])
        self.assertMatchesJsonBaseline(currdir + "load2a.out",
                                       currdir + "load2a.jsn")
        if yaml_available:
            pyutilib.subprocess.run([sys.executable, currdir + os.sep +
                                     "load2a.py", currdir, "yaml"])
            self.assertMatchesYamlBaseline(currdir + "load2a.out",
                                           currdir + "load2a.yml")


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['ExcelSpreadsheet']

import pyutilib.common
from pyutilib.excel.base import ExcelSpreadsheet_base

class _Interface(object):
    def __init__(self, available, module):
        self.available = available
        self.module = module

class Interfaces(object):
    singleton = None

    def __new__(cls):
        if Interfaces.singleton is None:
            Interfaces.singleton = super(Interfaces,cls).__new__(cls)
        return Interfaces.singleton

    def __init__(self):
        self.options = ['xlrd','win32com','openpyxl']
        self._modules = {}

        try:
            import xlrd
            from pyutilib.excel.spreadsheet_xlrd import (
                ExcelSpreadsheet_xlrd as module
            )
            self._modules['xlrd'] = _Interface(True, module)
        except ImportError:
            self._modules['xlrd'] = _Interface(False, None)

        try:
            from win32com.client.dynamic import Dispatch
            from pythoncom import CoInitialize, CoUninitialize
            from pythoncom import CoInitialize, CoUninitialize, com_error
            from pyutilib.excel.spreadsheet_win32com import (
                ExcelSpreadsheet_win32com as module
            )
            self._modules['win32com'] = _Interface(True, module)
        except ImportError:
            self._modules['win32com'] = _Interface(False, None)

        try:
            import openpyxl
            from pyutilib.excel.spreadsheet_openpyxl import (
                ExcelSpreadsheet_openpyxl as module
            )
            self._modules['openpyxl'] = _Interface(True, module)
        except ImportError:
            self._modules['openpyxl'] = _Interface(False, None)

    def __getitem__(self, item):
        return self._modules[item]


class ExcelSpreadsheet(ExcelSpreadsheet_base):

    def __new__(cls, *args, **kwds):
        #
        # Note that this class returns class instances rather than
        # class types.  This is because these classes are not
        # subclasses of ExcelSpreadsheet, and thus the __init__
        # method will not be called unless we construct the
        # class instances here.
        #
        ctype = kwds.pop('ctype', None)
        interfaces = Interfaces()

        if not ctype:
            for interface in interfaces.options:
                if interfaces[interface].available:
                    ctype = interface
                    break
            if not ctype:
                raise RuntimeError("No excel interface (from %s) available"
                                   % (interfaces.options,))

        if ctype not in interfaces.options:
            raise RuntimeError(
                "Excel interface %s not in known interfaces (%s)"
                % (ctype, interfaces.options,))

        if not interfaces[ctype].available:
            raise ImportError("Excel interface %s is not available" % (ctype,))

        return interfaces[ctype].module(*args, **kwds)



"""
Defining a Functor-specific task class
"""

__all__ = ['functor_api', 'IFunctorTask', 'FunctorAPIFactory', 'FunctorAPIData']

import inspect
import logging

from pyutilib.workflow.tasks import TaskPlugin
import pyutilib.component.core
import pyutilib.misc

logger = logging.getLogger('pyutilib.workflow')


class FunctorAPIData(dict):
    """
    A generalization of pyutilib.misc.Bunch.  This class counts access to attributes, and
    it generates errors for undefined attributes.
    """

    def __init__(self, **kw):
        dict.__init__(self, kw)
        self.__dict__.update(kw)
        self._declared_ = set()
        self.clean()

    def clean(self):
        self._dirty_ = set()

    def unused(self):
        for k, v in self.__dict__.items():
            if not k in self._dirty_:
                yield k

    def declare(self, args):
        if type(args) in (list, tuple, set):
            self._declared_.update(args)
        else:
            self._declared_.add(args)

    def update(self, d):
        self.__dict__.update(d)

    def __setitem__(self, name, val):
        self.__setattr__(name, val)

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __setattr__(self, name, val):
        if name[0] != '_':
            if len(self._declared_) > 0 and not name in self._declared_:
                raise AttributeError("Undeclared attribute '%s'" % name)
            self._dirty_.add(name)
            dict.__setitem__(self, name, val)
        self.__dict__[name] = val

    def __getattr__(self, name):
        if len(self._declared_) > 0 and not name in self._declared_:
            raise AttributeError("Undeclared attribute '%s'" % name)
        try:
            self._dirty_.add(name)
            return dict.__getitem__(self, name)
        except:
            if name[0] == '_':
                raise AttributeError("Unknown attribute %s" % name)
        return None

    def __repr__(self):
        return dict.__repr__(self)

    def __str__(self, nesting=0, indent=''):
        attrs = []
        indentation = indent + "    " * nesting
        for k, v in self.__dict__.items():
            if not k.startswith("_"):
                text = [indentation, k, ":"]
                if isinstance(v, FunctorAPIData):
                    text.append('\n')
                    text.append(v.__str__(nesting + 1))
                else:
                    text.append(' ' + str(v))
                attrs.append("".join(text))
        attrs.sort()
        return "\n".join(attrs)


class IFunctorTask(pyutilib.component.core.Interface):
    """Interface for Functor tasks"""


FunctorAPIFactory = pyutilib.component.core.CreatePluginFactory(IFunctorTask)


class FunctorTask(TaskPlugin):

    def __init__(self, *args, **kwargs):
        self._fn = kwargs.pop('fn', None)
        #
        TaskPlugin.__init__(self, *args, **kwargs)

    def execute(self, debug=False):
        if self._fn is None:  #pragma:nocover
            raise RuntimeError(
                "This is a bad definition of a FunctorTask.  The '_fn' method is not defined")
        #
        # Process data
        #
        data = self._kwds.get('data', None)
        if not data is None and type(data) is dict:
            _data = FunctorAPIData()
            _data.update(data)
            self._kwds['data'] = _data
        data = self._kwds.get('data', None)

        #
        # Test nested data
        #
        def nested_lookup(kwds, lookup):
            lookups = lookup.split('.')
            obj = kwds[lookups[0]]
            if not data is None and lookups[0] == 'data':
                data.declare(lookups[1])
            for key in lookups[1:]:
                #print key, obj
                obj = obj[key]
            if obj is None:
                raise ValueError

        for name in self._nested_requirements:
            try:
                nested_lookup(self._kwds, name)
            except ValueError:
                raise RuntimeError("None value found for nested attribute '%s'"
                                   % name)
            except:
                raise RuntimeError(
                    "Failed to verify existence of nested attribute '%s'" %
                    name)
        #
        # Call _fn
        #
        if 'data' in self._kwargs:
            data = self._kwds.get('data', None)
            retval = self._fn(**self._kwds)
        else:
            data = self._kwds.pop('data')
            retval = self._fn(data, **self._kwds)
        #
        # Process retval
        #
        if retval is None or id(data) == id(retval):
            self._retval = FunctorAPIData(data=data)
        elif isinstance(retval, FunctorAPIData):
            if not id(data) == id(retval):
                retval.data = data
            self._retval = retval
        elif isinstance(retval, dict):
            self._retval = FunctorAPIData()
            self._retval.update(retval)
        else:
            raise RuntimeError(
                "A Functor task function must return either None, a FunctorAPIData object, or an instance of dict.")

    def _call_start(self):
        self.reset()

    def _call_init(self, *options, **kwds):
        if not 'data' in kwds:
            if len(options) > 0:
                if len(options) > 1:
                    raise RuntimeError(
                        "A FunctorTask instance can only be executed with a single non-keyword argument")
                kwds['data'] = options[0]
                #options = options[1:]
            elif not self.inputs.data.optional:
                raise RuntimeError(
                    "A FunctorTask instance must be executed with at 'data' argument")
        self._kwds = kwds
        return TaskPlugin._call_init(self, **kwds)

    def _call_fini(self, *options, **kwds):
        for key in self._retval:
            if not key in self.outputs:
                raise RuntimeError(
                    "Cannot return value '%s' that is not a predefined output of a Functor task"
                    % key)
            setattr(self, key, self._retval[key])
        TaskPlugin._call_fini(self, *options, **kwds)
        return self._retval


#
# Decorate functions that are Functor tasks
#
def functor_api(fn=None, implements=None, outputs=None, namespace=None):

    def my_decorator(fn):
        if fn is None:  #pragma:nocover
            logger.error("Error applying decorator.  No function value!")
            return

        if namespace is None:
            _alias = fn.__name__
        else:
            _alias = namespace + '.' + fn.__name__
        _name = _alias.replace('_', '.')

        argspec = inspect.getargspec(fn)
        if not argspec.varargs is None:
            logger.error(
                "Attempting to declare Functor task with function '%s' that contains variable arguments"
                % _alias)
            return  #pragma:nocover
        if not argspec.keywords is None:
            logger.error(
                "Attempting to declare Functor task with function '%s' that contains variable keyword arguments"
                % _alias)
            return  #pragma:nocover

        if _alias in FunctorAPIFactory.services():
            logger.error(
                "Cannot define API %s, since this API name is already defined" %
                _alias)
            return  #pragma:nocover

        class TaskMeta(pyutilib.component.core.PluginMeta):

            def __new__(cls, name, bases, d):
                return pyutilib.component.core.PluginMeta.__new__(
                    cls, "FunctorTask_" + str(_name), bases, d)

        class FunctorTask_tmp(FunctorTask):

            __metaclass__ = TaskMeta

            pyutilib.component.core.alias(_alias)

            pyutilib.component.core.implements(IFunctorTask)

            def __init__(self, *args, **kwargs):
                kwargs['fn'] = fn
                FunctorTask.__init__(self, *args, **kwargs)
                if not fn is None:
                    if len(argspec.args) == 0:
                        nargs = 0
                    elif argspec.defaults is None:
                        nargs = len(argspec.args)
                    else:
                        nargs = len(argspec.args) - len(argspec.defaults)
                    self._kwargs = argspec.args[nargs:]
                    if nargs != 1 and 'data' not in self._kwargs:
                        logger.error("Functor '%s' must have a 'data argument" %
                                     _alias)
                    if argspec.defaults is None:
                        _defaults = {}
                    else:
                        _defaults = dict(
                            zip(argspec.args[nargs:], argspec.defaults))
                    #
                    docinfo = parse_docstring(fn)
                    #
                    if 'data' in docinfo['optional']:
                        self.inputs.declare(
                            'data',
                            doc='A container of labeled data.',
                            optional=True)
                    else:
                        self.inputs.declare(
                            'data', doc='A container of labeled data.')
                    for name in argspec.args[nargs:]:
                        if name in docinfo['optional']:
                            self.inputs.declare(
                                name,
                                optional=True,
                                default=_defaults[name],
                                doc=docinfo['optional'][name])
                        elif name in docinfo['required']:
                            self.inputs.declare(
                                name, doc=docinfo['required'][name])
                        elif name != 'data':
                            logger.error(
                                "Argument '%s' is not specified in the docstring!"
                                % name)
                    #
                    self.outputs.declare(
                        'data', doc='A container of labeled data.')
                    if outputs is None:
                        _outputs = list(docinfo['return'].keys())
                    else:
                        _outputs = outputs
                    for name in _outputs:
                        if name in docinfo['return']:
                            self.outputs.declare(
                                name, doc=docinfo['return'][name])
                        else:
                            logger.error(
                                "Return value '%s' is not specified in the docstring!"
                                % name)
                    #
                    self._nested_requirements = []
                    for name in docinfo['required']:
                        if '.' in name:
                            self._nested_requirements.append(name)
                    #
                    #  Error check keys for docinfo
                    #
                    for name in docinfo['required']:
                        if '.' in name:
                            continue
                        if not name in self.inputs:
                            logger.error(
                                "Unexpected name '%s' in list of required inputs for functor '%s'"
                                % (name, _alias))
                    for name in docinfo['optional']:
                        if not name in self.inputs:
                            logger.error(
                                "Unexpected name '%s' in list of optional inputs for functor '%s'"
                                % (name, _alias))
                    for name in docinfo['return']:
                        if not name in self.outputs:
                            logger.error(
                                "Unexpected name '%s' in list of outputs for functor '%s'"
                                % (name, _alias))
                    #
                    self.__help__ = fn.__doc__
                    self.__doc__ = fn.__doc__
                    self.__short_doc__ = docinfo['short_doc'].strip()
                    self.__long_doc__ = docinfo['long_doc'].strip()
                    self.__namespace__ = namespace

        return FunctorTask_tmp()

    if fn is None:
        return my_decorator
    return my_decorator(fn)


def parse_docstring(fn):
    """Parse a function docstring for information about the function arguments and return values"""
    retval = {}
    retval['short_doc'] = ""
    retval['long_doc'] = None
    retval['required'] = {}
    retval['optional'] = {}
    retval['return'] = {}
    curr = None
    doc = inspect.getdoc(fn)
    if doc is None:
        retval['long_doc'] = ''
        return retval
    for line in doc.split('\n'):
        line = line.strip()
        if line == 'Required:' or line == 'Required Arguments:':
            curr = 'required'
        elif line == 'Optional:' or line == 'Optimal Arguments:':
            curr = 'optional'
        elif line == 'Return Values:' or line == 'Return:' or line == 'Returned:':
            curr = 'return'
        elif curr is None:
            if retval['long_doc'] is None:
                if line == '':
                    retval['long_doc'] = ''
                else:
                    retval['short_doc'] += line
                    retval['short_doc'] += '\n'
            else:
                retval['long_doc'] += line
                retval['long_doc'] += '\n'
        elif ':' in line:
            name, desc = line.split(':', 1)
            retval[curr][name.strip()] = desc.strip() + '\n'
        else:
            retval[curr][name.strip()] += line + '\n'
    if retval['long_doc'] is None:
        retval['long_doc'] = ''
    return retval



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

__all__ = ['run', 'main', 'create_test_suites']

import sys
import optparse
import re
import os
from os.path import dirname, abspath

import pyutilib.th as unittest
from pyutilib.misc import Options
from pyutilib.component.core import ExtensionPoint
from pyutilib.autotest import plugins

# GAH: Inside create_test_suite all options are (were) being cast to
#      str(). Although I'm not sure, I assume this is to wash out
#      unicode coming from json files in Python 2.x. However, we don't
#      want to cast numeric options in either case (which was
#      happening, e.g., tolerance). This is my quick hack to fix
#      things. Please fix if you have a better understanding of what
#      should be happening.
try:
    unicode
except:
    basestring = str


def _str(x):
    if isinstance(x, basestring):
        return str(x)
    return x


def validate_test_config(suite):
    if suite is None:
        raise IOError(
            "Empty suite indicates problem processing suite configuration")
    #
    tmp = set(suite.keys())
    if not tmp.issubset(
            set(['python', 'solvers', 'problems', 'suites', 'driver'])):
        raise IOError("Unexpected test sections: " + str(suite.keys()))
    #
    if 'python' in suite:
        if not type(suite['python']) is list:
            raise IOError("Expected list of Python expressions")
    #
    if 'solvers' in suite:
        if not type(suite['solvers']) is dict:
            raise IOError("Expected dictionary of solvers")
        for key in suite['solvers']:
            if suite['solvers'][key] is None:
                suite['solvers'][key] = {}
            elif not type(suite['solvers'][key]) is dict:
                raise IOError(
                    "Expected solvers to have a dictionary of options: %s" %
                    str(suite))
    #
    if 'problems' in suite:
        if not type(suite['problems']) is dict:
            raise IOError("Expected dictionary of problems")
        for key in suite['problems']:
            if suite['problems'][key] is None:
                suite['problems'][key] = {}
            elif not type(suite['problems'][key]) is dict:
                raise IOError(
                    "Expected problems to have a dictionary of options")
    #
    if 'suites' in suite:
        if not type(suite['suites']) is dict:
            raise IOError("Expected dictionary of suites")
        for key in suite['suites']:
            if suite['suites'][key] is None:
                suite['suites'][key] = {}
            elif not type(suite['suites'][key]) is dict:
                raise IOError("Expected suites to have a dictionary of options")


@unittest.nottest
def create_test_suites(filename=None, config=None, _globals=None, options=None):
    if options is None:  #pragma:nocover
        options = Options()
    #
    # Add categories specified by the PYUTILIB_AUTOTEST_CATEGORIES
    # or PYUTILIB_UNITTEST_CATEGORIES environments
    #
    if options is None or options.categories is None or len(
            options.categories) == 0:
        options.categories = set()
        if 'PYUTILIB_AUTOTEST_CATEGORIES' in os.environ:
            for cat in re.split(',',
                                os.environ['PYUTILIB_AUTOTEST_CATEGORIES']):
                if cat != '':
                    options.categories.add(cat.strip())
        elif 'PYUTILIB_UNITTEST_CATEGORIES' in os.environ:
            for cat in re.split(',',
                                os.environ['PYUTILIB_UNITTEST_CATEGORIES']):
                if cat != '':
                    options.categories.add(cat.strip())
    #
    if not filename is None:
        if options.currdir is None:
            options.currdir = dirname(abspath(filename)) + os.sep
        #
        ep = ExtensionPoint(plugins.ITestParser)
        ftype = os.path.splitext(filename)[1]
        if not ftype == '':
            ftype = ftype[1:]
        service = ep.service(ftype)
        if service is None:
            raise IOError(
                "Unknown file type.  Cannot load test configuration from file '%s'"
                % filename)
        config = service.load_test_config(filename)
    #service.print_test_config(config)
    validate_test_config(config)
    #
    # Evaluate Python expressions
    #
    for item in config.get('python', []):
        try:
            exec(item, _globals)
        except Exception:
            err = sys.exc_info()[1]
            print("ERROR executing '%s'" % item)
            print("  Exception: %s" % str(err))
    #
    # Create test driver, which is put in the global namespace
    #
    driver = plugins.TestDriverFactory(config['driver'])
    if driver is None:
        raise IOError("Unexpected test driver '%s'" % config['driver'])
    _globals["test_driver"] = driver
    #
    # Generate suite
    #
    for suite in config.get('suites', {}):
        create_test_suite(suite, config, _globals, options)


@unittest.nottest
def create_test_suite(suite, config, _globals, options):
    #
    # Skip suite creation if the options categores do not intersect with the list of test suite categories
    #
    if len(options.categories) > 0:
        flag = False
        for cat in options.categories:
            if cat in config['suites'][suite].get('categories', []):
                flag = True
                break
        if not flag:
            return
    #
    # Create test driver
    #
    if suite in _globals:
        raise IOError(
            "Cannot create suite '%s' since there is another symbol with that name in the global namespace!"
            % suite)

    def setUpClassFn(cls):
        options = cls._options[None]
        cls._test_driver.setUpClass(cls, options)

    _globals[suite] = type(
        str(suite),
        (unittest.TestCase,), {'setUpClass': classmethod(setUpClassFn)})
    _globals[suite]._options[None] = options
    setattr(_globals[suite], '_test_driver', _globals['test_driver'])
    setattr(_globals[suite], 'suite_categories', config['suites'][suite].get(
        'categories', []))
    #
    # Create test functions
    #
    tests = []
    if 'tests' in config['suites'][suite]:
        for item in config['suites'][suite]['tests']:
            tests.append((item['solver'], item['problem'], item))
    else:
        for solver in config['suites'][suite]['solvers']:
            for problem in config['suites'][suite]['problems']:
                tests.append((solver, problem, {}))
    #
    for solver, problem, item in tests:
        ##sname = solver
        if options.testname_format is None:
            test_name = solver + "_" + problem
        else:
            test_name = options.testname_format % (solver, problem)
        #
        def fn(testcase, name, suite):
            options = testcase._options[suite, name]
            fn.test_driver.setUp(testcase, options)
            ans = fn.test_driver.run_test(testcase, name, options)
            fn.test_driver.tearDown(testcase, options)
            return ans

        fn.test_driver = _globals['test_driver']
        #
        _options = Options()
        #
        problem_options = config['suites'][suite]['problems'][problem]
        if not problem_options is None and 'problem' in problem_options:
            _problem = problem_options['problem']
        else:
            _problem = problem
        for attr, value in config['problems'].get(_problem, {}).items():
            _options[attr] = _str(value)
        if not problem_options is None:
            for attr, value in problem_options.items():
                _options[attr] = _str(value)
        #
        solver_options = config['suites'][suite]['solvers'][solver]
        if not solver_options is None and 'solver' in solver_options:
            _solver = solver_options['solver']
        else:
            _solver = solver
        _name = _solver
        for attr, value in config['solvers'].get(_solver, {}).items():
            _options[attr] = _str(value)
            if attr == 'name':
                _name = value
        if not solver_options is None:
            for attr, value in solver_options.items():
                _options[attr] = _str(value)
        #
        for key in item:
            if key not in ['problem', 'solver']:
                _options[key] = _str(item[key])
        #
        _options.solver = _str(_name)
        _options.problem = _str(_problem)
        _options.suite = _str(suite)
        _options.currdir = _str(options.currdir)
        #
        _globals[suite].add_fn_test(
            name=test_name, fn=fn, suite=suite, options=_options)


def cleanup(_globals, suites):
    for suite in suites:
        del _globals[suite]


def run(argv, _globals=None):
    #
    # Set sys.argv to the value specified by the user
    #
    sys.argv = argv
    #
    # Create the option parser
    #
    parser = optparse.OptionParser()
    parser.remove_option('-h')
    #
    parser.add_option(
        '-h',
        '--help',
        action='store_true',
        dest='help',
        default=False,
        help='Print command options')
    #
    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        dest='debug',
        default=False,
        help='Set debugging flag')
    #
    parser.add_option(
        '-v',
        '--verbose',
        action='store_true',
        dest='verbose',
        default=False,
        help='Verbose output')
    #
    parser.add_option(
        '-q',
        '--quiet',
        action='store_true',
        dest='quiet',
        default=False,
        help='Minimal output')
    #
    parser.add_option(
        '-f',
        '--failfast',
        action='store_true',
        dest='failfast',
        default=False,
        help='Stop on first failure')
    #
    parser.add_option(
        '-c',
        '--catch',
        action='store_true',
        dest='catch',
        default=False,
        help='Catch control-C and display results')
    #
    parser.add_option(
        '-b',
        '--buffer',
        action='store_true',
        dest='buffer',
        default=False,
        help='Buffer stdout and stderr durring test runs')
    #
    parser.add_option(
        '--cat',
        '--category',
        action='append',
        dest='categories',
        default=[],
        help='Define a list of categories that filter the execution of test suites')
    #
    parser.add_option(
        '--help-suites',
        action='store_true',
        dest='help_suites',
        default=False,
        help='Print the test suites that can be executed')
    #
    parser.add_option(
        '--help-tests',
        action='store',
        dest='help_tests',
        default=None,
        help='Print the tests in the specified test suite')
    #
    parser.add_option(
        '--help-categories',
        action='store_true',
        dest='help_categories',
        default=False,
        help='Print the test suite categories that can be specified')
    #
    # Parse the argument list and print help info if needed
    #
    _options, args = parser.parse_args(sys.argv)
    if _options.help:
        parser.print_help()

        print("""
Examples:
  %s                               - run all test suites
  %s MyTestCase.testSomething      - run MyTestCase.testSomething
  %s MyTestCase                    - run all 'test*' test methods
                                               in MyTestCase
""" % (args[0], args[0], args[0]))
        return
    #
    # If no value for _globals is specified, then we use the current context.
    #
    if _globals is None:
        _globals = globals()
    #
    # Setup and Options object and create test suites from the specified
    # configuration files.
    #
    options = Options()
    options.debug = _options.debug
    options.verbose = _options.verbose
    options.quiet = _options.quiet
    options.categories = _options.categories
    _argv = []
    for arg in args[1:]:
        if os.path.exists(arg):
            create_test_suites(filename=arg, _globals=_globals, options=options)
        else:
            _argv.append(arg)
    #
    # Collect information about the test suites:  suite names and categories
    #
    suites = []
    categories = set()
    for key in _globals.keys():
        if type(_globals[key]) is type and issubclass(_globals[key],
                                                      unittest.TestCase):
            suites.append(key)
            for c in _globals[key].suite_categories:
                categories.add(c)
    #
    # Process the --help-tests option
    #
    if _options.help_tests and not _globals is None:
        suite = _globals.get(_options.help_tests, None)
        if not type(suite) is type:
            print("Test suite '%s' not found!" % str(_options.help_tests))
            return cleanup(_globals, suites)
        tests = []
        for item in dir(suite):
            if item.startswith('test'):
                tests.append(item)
        print("")
        if len(tests) > 0:
            print("Tests defined in test suite '%s':" % _options.help_tests)
            for tmp in sorted(tests):
                print("    " + tmp)
        else:
            print("No tests defined in test suite '%s':" % _options.help_tests)
        print("")
        return cleanup(_globals, suites)
    #
    # Process the --help-suites and --help-categories options
    #
    if (_options.help_suites or
            _options.help_categories) and not _globals is None:
        if _options.help_suites:
            print("")
            if len(suites) > 0:
                print("Test suites defined in '%s':" %
                      os.path.basename(argv[0]))
                for suite in sorted(suites):
                    print("    " + suite)
            else:
                print("No test suites defined in '%s'!" %
                      os.path.basename(argv[0]))
            print("")
        if _options.help_categories:
            tmp = list(categories)
            print("")
            if len(tmp) > 0:
                print("Test suite categories defined in '%s':" %
                      os.path.basename(argv[0]))
                for c in sorted(tmp):
                    print("    " + c)
            else:
                print("No test suite categories defined in '%s':" %
                      os.path.basename(argv[0]))
            print("")
        return cleanup(_globals, suites)
    #
    # Reset the value of sys.argv per the expectations of the unittest module
    #
    tmp = [args[0]]
    if _options.quiet:
        tmp.append('-q')
    if _options.verbose or _options.debug:
        tmp.append('-v')
    if _options.failfast:
        tmp.append('-f')
    if _options.catch:
        tmp.append('-c')
    if _options.buffer:
        tmp.append('-b')
    tmp += _argv
    sys.argv = tmp
    #
    # Execute the unittest main function to run tests
    #
    unittest.main(module=_globals['__name__'])
    cleanup(_globals, suites)


def main(_globals=None):
    run(sys.argv, _globals=_globals)



#
# Unit Tests for pyutilib.options.configuration
#
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

from nose.tools import nottest
from pyutilib.component.core import ExtensionPoint, Plugin, PluginGlobals
from pyutilib.component.config.options import FileOption, declare_option
from pyutilib.component.config import Configuration, ConfigurationError
import pyutilib.th as unittest
import pyutilib.misc


def filter(line):
    return line.startswith(
        ";   section='") or 'memmon' in line or 'valgrind' in line or '[executables]' in line or 'null' in line


class Test(unittest.TestCase):

    class TMP(Plugin):

        def __init__(self):
            declare_option("a")
            declare_option("b", local_name="bb")
            declare_option("b")
            declare_option("c")
            declare_option("zz", section='a.b')
            declare_option("yy", default="foo")

    def setUp(self):
        PluginGlobals.add_env("testing.config")
        pyutilib.component.config.plugin_ConfigParser.Configuration_ConfigParser(
            name="Configuration_ConfigParser")
        self.tmp = Test.TMP()

    def tearDown(self):
        del self.tmp
        PluginGlobals.remove_env(
            "testing.config", cleanup=True, singleton=False)

    def test_init(self):
        """Test Configuration construction"""
        config = Configuration()

    def test_contains(self):
        """Test contains method"""
        config = Configuration()
        self.assertFalse("globals" in config)
        config.load(currdir + "config1.ini")
        self.assertTrue("globals" in config)

    def test_getitem(self):
        """Test getitem method"""
        config = Configuration()
        try:
            config["globals"]
            self.fail("expected error")
        except ConfigurationError:
            pass
        config.load(currdir + "config1.ini")
        keys = list(config["globals"].keys())
        keys.sort()
        self.assertTrue(keys == ["a", "b", "c"])

    def test_sections(self):
        """Test getitem method"""
        config = Configuration()
        config.load(currdir + "config1.ini")
        keys = list(config.sections())
        keys.sort()

    def test_load1(self):
        """Test load method"""
        config = Configuration()
        try:
            config.load(None)
            self.fail("expected error")
        except ConfigurationError:
            pass

    def test_load2(self):
        """Test load method"""
        config = Configuration()
        try:
            config.load("__missing__")
            self.fail("expected error")
        except ConfigurationError:
            pass

    def test_load3(self):
        """Test load method"""
        config = Configuration()
        try:
            config.load(currdir + "config2.ini")
            config.pprint()
            self.fail("expected error")
        except ConfigurationError:
            pass

    def test_load4(self):
        """Test load method"""
        config = Configuration()
        try:
            config.load(currdir + "config3.ini")
            self.fail("expected error")
        except ConfigurationError:
            pass

    def test_load5(self):
        """Test load method"""
        PluginGlobals.add_env("testing.config_loading")

        class TMP2(Plugin):

            def __init__(self):
                declare_option("a")
                declare_option("b", cls=FileOption)
                declare_option("c")
                declare_option("xx", section_re='globals.*')

        config = Configuration()
        tmp2 = TMP2()
        config.load(currdir + "config4.ini")
        #config.pprint()
        if False and sys.platform == "win32":
            #
            # A hack, to ensure cross-platform portability of this test
            #
            e = ExtensionPoint(IFileOption)
            for ep in e.extensions():
                ep.set_value("/dev/null", raw=True)
        #PluginGlobals.pprint()
        config.save(currdir + "config4.out")
        #print config
        self.assertFileEqualsBaseline(
            currdir + "config4.out", currdir + "config4.txt", filter=filter)
        pyutilib.misc.setup_redirect(currdir + "log2.out")
        config.pprint()
        pyutilib.misc.reset_redirect()
        self.assertFileEqualsBaseline(
            currdir + "log2.out", currdir + "log2.txt", filter=filter)
        PluginGlobals.remove_env(
            "testing.config_loading", cleanup=True, singleton=False)

    def test_save1(self):
        """Test save method"""
        config = Configuration()
        config.load(currdir + "config1.ini")
        #
        # A hack, to ensure cross-platform portability of this test
        #
        if False and sys.platform == "win32":
            e = ExtensionPoint(IFileOption)
            for ep in e.extensions():
                if ep.enabled():
                    ep.set_value("/dev/null", raw=True)
        config.save(currdir + "config1.out")
        #PluginGlobals.pprint()
        self.assertFileEqualsBaseline(
            currdir + "config1.out", currdir + "config1.txt", filter=filter)

    def test_save2(self):
        """Test save method"""
        config = Configuration()
        try:
            config.save(None)
            self.fail("expected error")
        except ConfigurationError:
            pass


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

# This software is adapted from the Trac software (specifically, the trac.core
# module.  The Trac copyright statement is included below.

__all__ = ['EggLoader']

import os
import sys
import logging
from pyutilib.component.config import ManagedPlugin
from pyutilib.component.core import implements, ExtensionPoint, IPluginLoader

logger = logging.getLogger('pyutilib.component.core.pca')
pkg_resources_avail = None

def _check_pkg_resources():
    # Defer import of pkg_resources until the EggLoader actually needs
    # it: loading pkg_resources is relatively slow, so this avoids the
    # time penalty every time pyutilib is imported.
    global pkg_resources_avail
    global pkg_resources
    try:
        if 'pkg_resources' not in sys.modules:
            #
            # Avoid a re-import, which causes setup.py warnings...
            #
            import pkg_resources
        else:
            pkg_resources = sys.modules['pkg_resources']
        pkg_resources_avail = True
    except ImportError:
        pkg_resources_avail = False

def pkg_environment(path):
    if pkg_resources_avail is None:
        _check_pkg_resources()
    return pkg_resources.Environment(path) if pkg_resources_avail else None

class EggLoader(ManagedPlugin):
    """
    Loader that looks for Python egg files in the plugins directories.
    These files get exampled with the pkg_resources package, and
    Plugin classes are loaded.

    Note: this plugin should not be directly constructed.  Instead,
    the user should employ the PluginFactory_EggLoader function.
    """

    implements(IPluginLoader, service=True)

    def __init__(self, **kwds):
        """EggLoader constructor.  The 'namespace' keyword option is
        required."""
        if 'name' not in kwds:
            kwds['name'] = "EggLoader." + kwds['namespace']
        super(EggLoader, self).__init__(**kwds)
        self.entry_point_name = kwds['namespace'] + ".plugins"
        if pkg_resources_avail is None:
            _check_pkg_resources()
        if not pkg_resources_avail:
            logger.warning(
                'A dummy EggLoader service is being constructed, because the pkg_resources package is not available on this machine.')

    def load(self, env, search_path, disable_re, name_re):
        generate_debug_messages = __debug__ and env.log.isEnabledFor(
            logging.DEBUG)
        if not pkg_resources_avail:
            if generate_debug_messages:
                env.log.debug(
                    'The EggLoader service is terminating early because the pkg_resources package is not available on this machine.')
            return

        working_set = pkg_resources.working_set

        env.log.info('BEGIN -  Loading plugins with an EggLoader service')
        distributions, errors = working_set.find_plugins(
            pkg_resources.Environment(search_path))
        for dist in distributions:
            if name_re.match(str(dist)):
                if generate_debug_messages:
                    env.log.debug('Adding plugin %r from %r', dist,
                                  dist.location)
                working_set.add(dist)
            else:
                if generate_debug_messages:
                    env.log.debug('Ignoring plugin %r from %r', dist,
                                  dist.location)

        def _log_error(item, e):
            gen_debug = __debug__ and env.log.isEnabledFor(logging.DEBUG)
            if isinstance(e, pkg_resources.DistributionNotFound):
                if gen_debug:
                    env.log.debug('Skipping "%s": ("%s" not found)', item, e)
            elif isinstance(e, pkg_resources.VersionConflict):
                if gen_debug:
                    env.log.debug('Skipping "%s": (version conflict "%s")',
                                  item, e)
            elif isinstance(e, pkg_resources.UnknownExtra):
                env.log.error('Skipping "%s": (unknown extra "%s")', item, e)
            elif isinstance(e, ImportError):
                env.log.error('Skipping "%s": (can\'t import "%s")', item, e)
            else:
                env.log.error('Skipping "%s": (error "%s")', item, e)

        for dist, e in errors.items():
            _log_error(dist, e)

        for entry in working_set.iter_entry_points(self.entry_point_name):
            if generate_debug_messages:
                env.log.debug('Loading %r from %r', entry.name,
                              entry.dist.location)
            try:
                entry.load(require=True)
            except (ImportError,
                    pkg_resources.DistributionNotFound,
                    pkg_resources.VersionConflict,
                    pkg_resources.UnknownExtra):
                e = sys.exc_info()[1]
                _log_error(entry, e)
            else:
                if not disable_re.match(os.path.dirname(
                        entry.module_name)) is None:
                    #_enable_plugin(env, entry.module_name)
                    pass

        env.log.info('END -    Loading plugins with an EggLoader service')

# Copyright (C) 2005-2008 Edgewall Software
# Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://trac.edgewall.org/log/.
#
# Author: Christopher Lenz <cmlenz@gmx.de>



#
# Unit Tests for util/misc
#
#

import pickle
import sys
import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep
import pyutilib.th as unittest
import pyutilib.misc
import pyutilib.misc.comparison


class Test(unittest.TestCase):

    def test_tostr(self):
        # Verify that tostr() generates a string
        str = pyutilib.misc.tostr([0.0, 1])
        self.assertEqual(str, "0.0 1")
        str = pyutilib.misc.tostr([])
        self.assertEqual(str, "")

    def test_flatten_tuple1(self):
        # Verify that flatten_tuple() flattens a normal tuple
        tmp = (1, "2", 3.0)
        ans = pyutilib.misc.flatten_tuple(tmp)
        self.assertEqual(ans, tmp)

    def test_flatten_tuple2(self):
        # Verify that flatten_tuple() flattens a nested tuple
        tmp = (1, "2", (4, ("5.0", (6))), 3.0)
        ans = pyutilib.misc.flatten_tuple(tmp)
        target = (1, "2", 4, "5.0", 6, 3.0)
        self.assertEqual(ans, target)

    def test_flatten_tuple3(self):
        # Verify that flatten_tuple() returns a non-tuple
        tmp = [1, "2", 3.0]
        ans = pyutilib.misc.flatten_tuple(tmp)
        self.assertEqual(ans, tmp)

    def test_flatten_tuple4(self):
        # Verify that flatten_tuple() removes empty tuples
        tmp = ((), 1, (), "2", ((), 4, ((), "5.0", (6), ()), ()), 3.0, ())
        ans = pyutilib.misc.flatten_tuple(tmp)
        target = (1, "2", 4, "5.0", 6, 3.0)
        self.assertEqual(ans, target)

    def test_flatten_tuple5(self):
        # Verify that flatten_tuple() can collapse to a single empty tuple
        self.assertEqual((1, 2, 3, 4, 5), pyutilib.misc.flatten_tuple((
            (), 1, (), 2, ((), 3, ((), 4, ()), ()), 5, ())))
        self.assertEqual((), pyutilib.misc.flatten_tuple(((((), ()), ()), ())))
        self.assertEqual((), pyutilib.misc.flatten_tuple(((), ((), ((),)))))

    def test_flatten_list1(self):
        # Verify that flatten_list() flattens a normal list
        tmp = [1, "2", 3.0]
        ans = pyutilib.misc.flatten_list(tmp)
        self.assertEqual(ans, tmp)

    def test_flatten_list2(self):
        # Verify that flatten_list() flattens a nested list
        tmp = [1, "2", [4, ["5.0", [6]]], 3.0]
        ans = pyutilib.misc.flatten_list(tmp)
        target = [1, "2", 4, "5.0", 6, 3.0]
        self.assertEqual(ans, target)

    def test_flatten_list3(self):
        # Verify that flatten_list() returns a non-list
        tmp = (1, "2", 3.0)
        ans = pyutilib.misc.flatten_list(tmp)
        self.assertEqual(ans, tmp)

    def test_flatten_list4(self):
        # Verify that flatten_list() removes empty lists
        tmp = [[], 1, [], "2", [[], 4, [[], "5.0", [6], []], []], 3.0, []]
        ans = pyutilib.misc.flatten_list(tmp)
        target = [1, "2", 4, "5.0", 6, 3.0]
        self.assertEqual(ans, target)

    def test_flatten_list5(self):
        # Verify that flatten_list() can collapse to a single empty list
        self.assertEqual([1, 2, 3, 4, 5], pyutilib.misc.flatten_list(
            [[], 1, [], 2, [[], 3, [[], 4, []], []], 5, []]))
        self.assertEqual([], pyutilib.misc.flatten_list([[[[], []], []], []]))
        self.assertEqual([], pyutilib.misc.flatten_list([[], [[], [[],]]]))

    def test_Bunch(self):
        a = 1
        b = "b"
        tmp = pyutilib.misc.Bunch(a=a, b=b)
        self.assertEqual(tmp.a, a)
        self.assertEqual(tmp.b, b)

    def test_Container1(self):
        opt = pyutilib.misc.Container('a=None c=d e="1 2 3"', foo=1, bar='x')
        self.assertEqual(opt.ll, None)
        self.assertEqual(opt.a, None)
        self.assertEqual(opt.c, 'd')
        self.assertEqual(opt.e, '1 2 3')
        self.assertEqual(opt.foo, 1)
        self.assertEqual(opt.bar, 'x')
        self.assertEqual(opt['bar'], 'x')
        opt.xx = 1
        opt['yy'] = 2
        self.assertEqual(
            set(opt.keys()), set(['a', 'bar', 'c', 'foo', 'e', 'xx', 'yy']))
        opt.x = pyutilib.misc.Container(a=1, b=2)
        self.assertEqual(
            set(opt.keys()), set(
                ['a', 'bar', 'c', 'foo', 'e', 'xx', 'yy', 'x']))
        self.assertEqual(
            repr(opt),
            "Container(a = None, bar = 'x', c = 'd', e = '1 2 3', foo = 1, x = Container(a = 1, b = 2), xx = 1, yy = 2)")
        self.assertEqual(
            str(opt), """a: None
bar: 'x'
c: 'd'
e: '1 2 3'
foo: 1
x:
    a: 1
    b: 2
xx: 1
yy: 2""")
        opt._name_ = 'CONTAINER'
        self.assertEqual(
            set(opt.keys()), set(
                ['a', 'bar', 'c', 'foo', 'e', 'xx', 'yy', 'x']))
        self.assertEqual(
            repr(opt),
            "Container(a = None, bar = 'x', c = 'd', e = '1 2 3', foo = 1, x = Container(a = 1, b = 2), xx = 1, yy = 2)")
        self.assertEqual(
            str(opt), """a: None
bar: 'x'
c: 'd'
e: '1 2 3'
foo: 1
x:
    a: 1
    b: 2
xx: 1
yy: 2""")

    def test_Container2(self):
        o1 = pyutilib.misc.Container('a=None c=d e="1 2 3"', foo=1, bar='x')
        s = pickle.dumps(o1)
        o2 = pickle.loads(s)
        self.assertEqual(o1, o2)

    def test_flatten1(self):
        # Test that flatten works correctly
        self.assertEqual("abc", pyutilib.misc.flatten("abc"))
        self.assertEqual(1, pyutilib.misc.flatten(1))
        self.assertEqual([1, 2, 3], pyutilib.misc.flatten((1, 2, 3)))
        self.assertEqual([1, 2, 3], pyutilib.misc.flatten([1, 2, 3]))
        self.assertEqual([1, 2, 3, 4], pyutilib.misc.flatten((1, 2, [3, 4])))
        self.assertEqual([1, 2, 'abc'], pyutilib.misc.flatten((1, 2, 'abc')))
        self.assertEqual([1, 2, 'abc'], pyutilib.misc.flatten((1, 2, ('abc',))))
        a = [0, 9, 8]
        self.assertEqual([1, 2, 0, 9, 8], pyutilib.misc.flatten((1, 2, a)))
        self.assertEqual([1, 2, 3, 4, 5], pyutilib.misc.flatten(
            [[], 1, [], 2, [[], 3, [[], 4, []], []], 5, []]))
        self.assertEqual([], pyutilib.misc.flatten([[[[], []], []], []]))
        self.assertEqual([], pyutilib.misc.flatten([[], [[], [[],]]]))

    def test_quote_split(self):
        ans = pyutilib.misc.quote_split("[ ]+", "a bb ccc")
        self.assertEqual(ans, ["a", "bb", "ccc"])
        ans = pyutilib.misc.quote_split("[ ]+", "")
        self.assertEqual(ans, [""])
        ans = pyutilib.misc.quote_split("[ ]+", 'a "bb ccc"')
        self.assertEqual(ans, ["a", "\"bb ccc\""])
        ans = pyutilib.misc.quote_split("[ ]+", "a 'bb ccc'")
        self.assertEqual(ans, ["a", "'bb ccc'"])
        ans = pyutilib.misc.quote_split("[ ]+", "a X\"bb ccc\"Y")
        self.assertEqual(ans, ["a", "X\"bb ccc\"Y"])
        ans = pyutilib.misc.quote_split("[ ]+", "a X'bb ccc'Y")
        self.assertEqual(ans, ["a", "X'bb ccc'Y"])
        ans = pyutilib.misc.quote_split("[ ]+", "a X'bb ccc'Y A")
        self.assertEqual(ans, ["a", "X'bb ccc'Y", "A"])
        try:
            ans = pyutilib.misc.quote_split("[ ]+", 'a "bb ccc')
            self.fail(
                "test_quote_split - failed to detect unterminated quotation")
        except ValueError:
            pass

        ans = pyutilib.misc.quote_split("a bb\\\" ccc")
        self.assertEqual(ans, ["a", "bb\\\"", "ccc"])
        self.assertRaises(ValueError, pyutilib.misc.quote_split,
                          ("a bb\\\\\" ccc"))
        ans = pyutilib.misc.quote_split("a \"bb  ccc\"")
        self.assertEqual(ans, ["a", "\"bb  ccc\""])
        ans = pyutilib.misc.quote_split("a 'bb \" ccc'")
        self.assertEqual(ans, ["a", "'bb \" ccc'"])
        ans = pyutilib.misc.quote_split("a \"bb ' ccc\"")
        self.assertEqual(ans, ["a", "\"bb ' ccc\""])
        ans = pyutilib.misc.quote_split("a \"bb \\\\\\\" ccc\"")
        self.assertEqual(ans, ["a", "\"bb \\\\\\\" ccc\""])
        ans = pyutilib.misc.quote_split('b', "abbbccc")
        self.assertEqual(ans, ["a", '', '', 'ccc'])
        ans = pyutilib.misc.quote_split('b+', "abbbccc")
        self.assertEqual(ans, ["a", 'ccc'])
        ans = pyutilib.misc.quote_split(' ', "a b\ c")
        self.assertEqual(ans, ["a", 'b\ c'])

    def test_tuplize(self):
        ans = pyutilib.misc.tuplize([0, 1, 2, 3, 4, 5], 2, "a")
        self.assertEqual(ans, [(0, 1), (2, 3), (4, 5)])
        try:
            ans = pyutilib.misc.tuplize([0, 1, 2, 3, 4, 5, 6], 2, "a")
            self.fail("test_tuplize failed to detect bad list length")
        except ValueError:
            pass

    def test_rmtree(self):
        if os.path.exists(".test_misc"):
            pyutilib.misc.rmtree(".test_misc")
        os.makedirs(".test_misc/a/b/c")
        OUTPUT = open(".test_misc/a/file", "w")
        OUTPUT.write("HERE\n")
        OUTPUT.close()
        pyutilib.misc.rmtree(".test_misc")
        if os.path.exists(".test_misc"):
            self.fail("test_rmtree failed to delete .test_misc dir")

    def test_search_file(self):
        # Test that search_file works
        ans = pyutilib.misc.search_file("foobar")
        self.assertEqual(ans, None)
        path = sys.path + [currdir]
        ans = pyutilib.misc.search_file("test1.cfg", search_path=path)
        self.assertEqual(ans, abspath(currdir + "test1.cfg"))
        ans = pyutilib.misc.search_file("test1.cfg", search_path=currdir)
        self.assertEqual(ans, abspath(currdir + "test1.cfg"))
        ans = pyutilib.misc.search_file(
            "test1", implicitExt=".cfg", search_path=path)
        self.assertEqual(ans, abspath(currdir + "test1.cfg"))

    def test_search_file2(self):
        # Test that search_file works with an empty PATH environment
        tmp = os.environ["PATH"]
        del os.environ["PATH"]
        ans = pyutilib.misc.search_file("_cd_")
        os.environ["PATH"] = tmp
        self.assertEqual(ans, None)

    def test_search_file3(self):
        # Test that search_file works with a validation function
        def validate_ls(filename):
            if filename.endswith('ls'):
                return False
            return True

        tmp = os.environ["PATH"]
        del os.environ["PATH"]
        ans = pyutilib.misc.search_file("ls", validate=validate_ls)
        os.environ["PATH"] = tmp
        self.assertEqual(ans, None)

    def test_file_compare1(self):
        # Test that file comparison works
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp1.txt", currdir + "filecmp1.txt")
        if flag:
            self.fail(
                "test_file_compare1 - found differences in filecmp1.txt at line "
                + str(lineno))
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp1.txt", currdir + "filecmp2.txt")
        if flag:
            self.fail(
                "test_file_compare1 - found differences between filecmp1.txt filecmp2.txt at line "
                + str(lineno))
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp1.txt", currdir + "filecmp3.txt")
        if not flag or lineno != 4:
            self.fail("test_file_compare1 - expected difference at line 4")
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp1.txt", currdir + "filecmp4.txt")
        if not flag or lineno != 3:
            self.fail("test_file_compare1 - expected difference at line 3")
        try:
            [flag, lineno, diffstr] = pyutilib.misc.compare_file(
                currdir + "foo.txt", currdir + "bar.txt")
            self.fail("test_file_compare1 - should have failed to find foo.txt")
        except IOError:
            pass
        try:
            [flag, lineno, diffstr] = pyutilib.misc.compare_file(
                currdir + "filecmp1.txt", currdir + "bar.txt")
            self.fail("test_file_compare1 - should have failed to find bar.txt")
        except IOError:
            pass

    def test_file_compare1a(self):
        # Test that file comparison works without numeric values
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp1.txt", currdir + "filecmp1.txt")
        if flag:
            self.fail(
                "test_file_compare1a - found differences in filecmp1.txt at line "
                + str(lineno))
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp1.txt", currdir + "filecmp2.txt")
        if flag:
            self.fail(
                "test_file_compare1a - found differences between filecmp1.txt filecmp2.txt at line "
                + str(lineno))
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp1.txt", currdir + "filecmp3.txt")
        if not flag or lineno != 4:
            self.fail("test_file_compare1a - expected difference at line 4",
                      ", got %s, %s" % (flag, lineno))
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp1.txt", currdir + "filecmp4.txt")
        if not flag or lineno != 3:
            self.fail("test_file_compare1a - expected difference at line 3"
                      ", got %s, %s" % (flag, lineno))
        try:
            [flag, lineno, diffstr
            ] = pyutilib.misc.compare_file_with_numeric_values(
                currdir + "foo.txt", currdir + "bar.txt")
            self.fail(
                "test_file_compare1a - should have failed to find foo.txt")
        except IOError:
            pass
        try:
            [flag, lineno, diffstr
            ] = pyutilib.misc.compare_file_with_numeric_values(
                currdir + "filecmp1.txt", currdir + "bar.txt")
            self.fail(
                "test_file_compare1a - should have failed to find bar.txt")
        except IOError:
            pass

    def test_file_compare1b(self):
        # Test that file comparison works with numeric values
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp6.txt", currdir + "filecmp7.txt")
        if not flag:
            self.fail(
                "test_file_compare1b - expected differences in filecmp6.txt and filecmp7.txt at line 1")
        #
        #
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp6.txt", currdir + "filecmp7.txt")
        if flag:
            self.fail(
                "test_file_compare1b - unexpected differences in filecmp6.txt and filecmp7.txt at line %d"
                % lineno)
        #
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp6.txt", currdir + "filecmp8.txt")
        if not flag:
            self.fail(
                "test_file_compare1b - expected differences in filecmp6.txt and filecmp8.txt at line 1")
        #
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp6.txt", currdir + "filecmp8.txt", tolerance=1e-2)
        if flag:
            self.fail(
                "test_file_compare1b - unexpected differences in filecmp6.txt and filecmp8.txt at line %d"
                % lineno)
        #
        [flag, lineno, diffstr
        ] = pyutilib.misc.compare_file_with_numeric_values(
            currdir + "filecmp10.txt",
            currdir + "filecmp11.txt",
            tolerance=1e-2)
        if flag:
            self.fail(
                "test_file_compare1b - unexpected differences in filecmp10.txt and filecmp11.txt at line %d"
                % lineno)
        #
        #
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp6.txt", currdir + "filecmp7.txt", tolerance=0.0)
        if flag:
            self.fail(
                "test_file_compare1b - unexpected differences in filecmp6.txt and filecmp7.txt at line %d"
                % lineno)
        #
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp6.txt", currdir + "filecmp8.txt", tolerance=0.0)
        if not flag:
            self.fail(
                "test_file_compare1b - expected differences in filecmp6.txt and filecmp8.txt at line 1")
        #
        [flag, lineno, diffstr] = pyutilib.misc.compare_file(
            currdir + "filecmp6.txt", currdir + "filecmp8.txt", tolerance=1e-2)
        if flag:
            self.fail(
                "test_file_compare1b - unexpected differences in filecmp6.txt and filecmp8.txt at line %d"
                % lineno)

    def test_file_compare2(self):
        # Test that large file comparison works
        flag = pyutilib.misc.compare_large_file(currdir + "filecmp1.txt",
                                                currdir + "filecmp1.txt")
        if flag:
            self.fail("test_file_compare2 - found differences in filecmp1.txt")
        flag = pyutilib.misc.compare_large_file(currdir + "filecmp1.txt",
                                                currdir + "filecmp2.txt")
        if flag:
            self.fail(
                "test_file_compare2 - found differences between filecmp1.txt filecmp2.txt")
        flag = pyutilib.misc.compare_large_file(
            currdir + "filecmp2.txt", currdir + "filecmp3.txt", bufSize=7)
        if not flag:
            self.fail(
                "test_file_compare2 - found differences between filecmp1.txt filecmp2.txt")
        flag = pyutilib.misc.compare_large_file(currdir + "filecmp1.txt",
                                                currdir + "filecmp3.txt")
        if not flag:
            self.fail("test_file_compare2 - expected difference")
        flag = pyutilib.misc.compare_large_file(currdir + "filecmp1.txt",
                                                currdir + "filecmp4.txt")
        if not flag:
            self.fail("test_file_compare2 - expected difference")
        try:
            flag = pyutilib.misc.compare_large_file(currdir + "foo.txt",
                                                    currdir + "bar.txt")
            self.fail("test_file_compare2 - should have failed to find foo.txt")
        except IOError:
            pass
        try:
            flag = pyutilib.misc.compare_large_file(currdir + "filecmp1.txt",
                                                    currdir + "bar.txt")
            self.fail("test_file_compare1 - should have failed to find bar.txt")
        except IOError:
            pass

    def test_remove_chars(self):
        # Test the remove_chars_in_list works
        a = pyutilib.misc.comparison.remove_chars_in_list("", "")
        self.assertEqual(a, "")
        a = pyutilib.misc.comparison.remove_chars_in_list("abcde", "")
        self.assertEqual(a, "abcde")
        a = pyutilib.misc.comparison.remove_chars_in_list("abcde", "ace")
        self.assertEqual(a, "bd")

    def test_get_desired_chars_from_file(self):
        # Test that get_desired_chars_from_file works
        INPUT = open(currdir + "filecmp5.txt", "r")
        a = pyutilib.misc.comparison.get_desired_chars_from_file(INPUT, 3,
                                                                 "b,d")
        self.assertEqual(a, "ace")
        INPUT.close()
        INPUT = open(currdir + "filecmp5.txt", "r")
        a = pyutilib.misc.comparison.get_desired_chars_from_file(INPUT, 100)
        self.assertEqual(a, "abcde\nfghij\n")
        INPUT.close()

    def test_sort_index1(self):
        # Test that sort_index returns the correct value for a sorted array
        ans = pyutilib.misc.sort_index(range(0, 10))
        self.assertEqual(ans, list(range(0, 10)))

    def test_sort_index2(self):
        # Test that sort_index returns an array that can be used to sort the data
        data = [4, 2, 6, 8, 1, 9, 3, 10, 7, 5]
        ans = pyutilib.misc.sort_index(data)
        sorted = []
        for i in range(0, len(data)):
            sorted.append(data[ans[i]])
        data.sort()
        self.assertEqual(data, sorted)

    def test_create_hardlink(self):
        orig = currdir + "import1.txt"
        link = currdir + "import1.txt.hardlink"

        try:
            os.remove(link)
        except OSError:
            pass
        self.assertTrue(os.path.exists(orig))
        self.assertTrue(not os.path.exists(link))
        pyutilib.misc.create_hardlink(orig, link)
        self.assertTrue(os.path.exists(link))
        self.assertTrue(os.path.isfile(link))
        os.remove(link)


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

import sys

if 'nose' not in sys.modules and 'nose2' not in sys.modules:
    raise ImportError("""pyutilib.enum has been removed.

Python 3 now has an enum implementation in the standard library (also
available for older Python versions as the third-party enum34 PyPI
package) that supersedes this library.""")



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
A class for interacting with an Excel spreadsheet.
"""

#
# Imports
#
import os
import warnings
import openpyxl

from pyutilib.excel.base import ExcelSpreadsheet_base


class ExcelSpreadsheet_openpyxl(ExcelSpreadsheet_base):

    def can_read(self):
        return True

    def can_write(self):
        return True

    def can_calculate(self):
        return False

    def __init__(self, filename=None, worksheets=(1,), default_worksheet=1):
        """
        Constructor.
        """
        self.wb = None
        self.xlsfile = None
        self._ws = {}
        if filename is not None:
            self.open(filename, worksheets, default_worksheet)

    def open(self, filename, worksheets=(1,), default_worksheet=1):
        """
        Initialize this object from a file.
        """
        #
        # Set the excel spreadsheet name
        #
        self.xlsfile = filename
        #
        # Start the excel spreadsheet
        #
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.wb = openpyxl.load_workbook(self.xlsfile)
        self.worksheets = self.wb.sheetnames
        self._ws = {}
        for wsid in worksheets:
            self._ws[wsid] = self.wb[self.worksheets[wsid - 1]]
        self.default_worksheet = default_worksheet

    def ws(self):
        """ The active worksheet """
        return self._ws[self.default_worksheet]

    def __del__(self):
        """
        Close the spreadsheet when deleting this object.
        """
        self.close()

    def activate(self, name):
        """
        Activate a specific sheet
        """
        if name is None:
            return
        if name in self.worksheets:
            idx = self.worksheets.index(name)
            self.wb.active = idx
        elif not name in self._ws:
            self.wb.add_sheet(name)
            idx = len(self.wb.get_sheet_names()) - 1
            self._ws[idx] = self.wb.Worksheets.Item(name)
            self._ws[idx].Activate()
        self.default_worksheet = idx

    def close(self):
        """
        Close the spreadsheet
        """
        if self is None:  #pragma:nocover
            return
        if self._ws is None:
            return
        self._ws = {}

    def calc_iterations(self, val=None):
        if val is None:
            return self.xl.Iteration
        if not type(val) is bool:
            raise ValueError(
                "ExcelSpreadsheet calc_iterations can only be set to a boolean")
        self.xl.Iteration = val

    def max_iterations(self, val=None):
        if val is None:
            return self.xl.MaxIterations
        if (not type(val) in [int, float, long]) or val < 0:
            raise ValueError(
                "ExcelSpreadsheet max_iterations can only be set to nonnegative integer")
        self.xl.MaxIterations = val

    def calculate(self):
        """
        Perform calculations in a spreadsheet
        """
        self.xl.Calculate()

    def set_array(self, row, col, val, wsid=None):
        self.activate(wsid)
        ws = self.wb.active
        xlo = row
        xhi = row + len(val) - 1
        ylo = col
        yhi = col + 1
        for row in range(xhi - xlo + 1):
            for col in range(yhi - ylo + 1):
                #print(row,col)
                ws.cell(row=xlo + row, column=ylo + col).value = val[row][col]

    def get_array(self, row, col, row2, col2, wsid=None, raw=False):
        self.activate(wsid)
        ws = self.wb.active
        ans = []
        for row in ws.iter_rows(min_col=col, min_row=row, max_col=col2, max_row=row2):
            ans_ = []
            for col in row:
                ans_.append(col.value)
            if len(ans_) == 1:
                ans.append(ans_[0])
            else:
                ans.append(list(ans_))
        return list(ans)

    def set_range(self, rangename, val, wsid=None):
        """
        Set a range with a given value (or set of values)
        """
        #
        # Process data into tuples
        #
        if type(val) in (int, float):
            val = ((val,),)
        if type(val) is tuple:
            data = val
        elif type(val) not in (float, int, bool):
            data = []
            for item in val:
                if type(item) is tuple:
                    data.append(item)
                elif type(item) in (float, int, bool):
                    data.append((item,))
                else:
                    data.append(tuple(item))
            data = tuple(data)
        #
        # Set the data
        #
        self.activate(wsid)
        _range = self._range(rangename)
        if not _range is None and len(val) != self.get_range_nrows(rangename):
            raise IOError("Setting data with " + str(len(val)) +
                          " rows but range has " + str(
                              self.get_range_nrows(rangename)))
        _destinations = list(_range.destinations)
        ws = self.wb[_destinations[0][0]]
        #
        ylo, xlo, yhi, xhi = openpyxl.worksheet.worksheet.range_boundaries(
            _destinations[0][1])
        for row in range(xhi - xlo + 1):
            for col in range(yhi - ylo + 1):
                ws.cell(row=xlo + row, column=ylo + col).value = data[row][col]

    def get_column(self, colname, wsid=None, raw=False, contiguous=False):
        """
        Select the values of a column.
        This ignores blank cells at the top and bottom of the column.

        If contiguous is False, a list is returned with all cell values
        starting from the first non-blank cell until the last non-blank cell.

        If contiguous if True, a list is returned with all cell values
        starting from the first non-blank cell until the first blank cell.
        """
        self.activate(wsid)
        name = colname + "1"
        if self.get_range(name) is None:
            start = self.ws().Range(name).End(self.xlDown)
        else:
            start = self.ws().Range(name)
        if contiguous:
            range = self.ws().Range(start, start.End(self.xlDown))
        else:
            range = self.ws().Range(
                start, self.ws().Range(colname + "65536").End(self.xlUp))
        tmp = self._get_range_data(range, raw)
        return tmp

    def get_range(self, rangename, wsid=None, raw=False):
        """
        Get values for a specified range
        """
        self.activate(wsid)
        _range = self._range(rangename)
        data = self._get_range_data(_range, raw)
        if len(data) == 1:
            return data[0]
        return data

    def _get_range_data(self, _range, raw):
        ans = []
        if type(_range) is tuple:
            _data = _range
        else:
            _destinations = list(_range.destinations)
            ws = self.wb[_destinations[0][0]]
            #
            # If we have a since cell, just return its value
            #
            if not ':' in _destinations[0][1]:
                return [ ws[_destinations[0][1]].value ]
            #
            # If we have a range, return a list of values
            #
            _data = ws[_destinations[0][1]]
        #
        # Process Data
        #
        for row in _data:
            rvals = []
            for cell in row:
                rvals.append(cell.value)
            if len(rvals) == 1:
                ans.append(rvals[0])
            else:
                ans.append(list(rvals))
        return list(ans)

    def get_range_nrows(self, rangename, wsid=None):
        """
        Get the number of rows in a specified range
        """
        self.activate(wsid)
        _range = self._range(rangename)
        data = self._get_range_data(_range, False)
        return len(data)

    def get_range_ncolumns(self, rangename, wsid=None):
        """
        Get the number of columns in a specified range
        """
        self.activate(wsid)
        _range = self._range(rangename)
        data = self._get_range_data(_range, False)
        if type(data[0]) is tuple:
            return len(data[0])
        return 1

    def _range(self, rangeid, wsid=None, exception=True):
        """
        Return a range for a given worksheet
        """
        self.activate(wsid)
        try:
            #
            # If rangeid is a tuple, then this is a list of arguments 
            #
            if type(rangeid) is tuple:
                return self.wb.get_squared_range(*rangeid)
            #
            # Otherwise, we assume that this is a range name.
            #
            try:
                return self.wb.defined_names[rangeid]
            except KeyError:
                pass
            ws = self.wb.active
            if ':' in rangeid:
                _rangeid = rangeid.split(':')
                return ws[_rangeid[0]:_rangeid[1]]
            else:
                return ws[rangeid:rangeid]
        except:
            raise IOError("Unknown range name `" + str(rangeid) + "'")



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________


def add_method(self, method, name=None):
    """
    Add a method to a class:

      self   - the object that is modified
      method - a function
      name   - a string that describes the new method name

    Adapted from code submitted by Moshe Zadka to the
    ActiveState Programmer Network http://aspn.activestate.com
    """
    if name is None:
        name = method.__name__

    class new(self.__class__):
        pass

    setattr(new, name, method)
    self.__class__ = new
    return getattr(self, name)


def add_method_by_name(self,
                       method_name,
                       name=None,
                       globals=globals(),
                       locals=None):
    """
    Add a method to a class given a function class

      self        - the object that is modified
      method_name - the name of a function
      name        - a string that describes the new method name
      globals     - the global dictionary
      locals      - the local dictionary

    Adapted from code submitted by Moshe Zadka to the
    ActiveState Programmer Network http://aspn.activestate.com
    """
    add_method(self, eval(method_name, globals, locals), name)
